"""Tests for thermal property calculation."""

import numpy as np

temps = [
    0.000000,
    100.000000,
    200.000000,
    300.000000,
    400.000000,
    500.000000,
    600.000000,
    700.000000,
    800.000000,
    900.000000,
]
fes = [
    0.000000,
    3.094415,
    -0.699291,
    -7.092875,
    -15.174823,
    -24.512227,
    -34.850111,
    -46.019993,
    -57.902128,
    -70.406982,
]
entropies = [
    0.000000,
    18.743138,
    53.183741,
    73.330203,
    87.624345,
    98.711742,
    107.770806,
    115.430135,
    122.064948,
    127.917267,
]
cvs = [
    0.000000,
    49.687287,
    49.687287,
    49.687287,
    49.687287,
    49.687287,
    49.687287,
    49.687287,
    49.687287,
    49.687287,
]


def test_thermal_properties(ph_nacl):
    """Test thermal property calculation with t_step and t_max parameters."""
    ph_nacl.run_mesh([5, 5, 5])
    ph_nacl.run_thermal_properties(
        t_step=100, t_max=900, cutoff_frequency=1e-5, classical=True
    )
    _test_thermal_properties(ph_nacl)


def test_thermal_properties_at_temperatues(ph_nacl):
    """Test thermal property calculation with temperatures parameter."""
    ph_nacl.run_mesh([5, 5, 5])
    temperatures = [0, 100, 200, 300, 400, 500, 600, 700, 800, 900]
    ph_nacl.run_thermal_properties(
        temperatures=temperatures, cutoff_frequency=1e-5, classical=True
    )
    _test_thermal_properties(ph_nacl)


def _test_thermal_properties(ph):
    tp = ph.thermal_properties

    for vals in tp.thermal_properties:
        print(",\n".join(["%.6f" % v for v in vals]))

    for i in range(2):
        if i == 1:
            tp.run(lang="py")
        for vals_ref, vals in zip((temps, fes, entropies, cvs), tp.thermal_properties):
            np.testing.assert_allclose(vals_ref, vals, atol=1e-5)



"""Core routine of linear tetrahedon method on regular grid."""

# Copyright (C) 2013 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings

import numpy as np


def get_tetrahedra_relative_grid_address(microzone_lattice, lang="C"):
    """Return relative (differences of) grid addresses from the central.

    Parameter
    ---------
    microzone_lattice : ndarray or list of list
        column vectors of parallel piped microzone lattice, i.e.,
        microzone_lattice = np.linalg.inv(cell.get_cell()) / mesh

    """
    try:
        import phonopy._phonopy as phonoc
    except ImportError:
        import sys

        print("Phonopy C-extension has to be built properly.")
        sys.exit(1)

    relative_grid_address = np.zeros((24, 4, 3), dtype="int64", order="C")
    if lang == "C":
        phonoc.tetrahedra_relative_grid_address(
            relative_grid_address,
            np.array(microzone_lattice, dtype="double", order="C"),
        )
    else:
        relative_grid_address = _get_relative_grid_addresses_from_microzone_lattice(
            microzone_lattice
        )[0]
    return relative_grid_address


def get_all_tetrahedra_relative_grid_address(lang="C"):
    """Return relative grid addresses dataset.

    This exists only for the test.

    """
    try:
        import phonopy._phonopy as phonoc
    except ImportError:
        import sys

        print("Phonopy C-extension has to be built properly.")
        sys.exit(1)

    relative_grid_address = np.zeros((4, 24, 4, 3), dtype="int64", order="C")
    if lang == "C":
        phonoc.all_tetrahedra_relative_grid_address(relative_grid_address)
    else:
        for i in range(4):
            relative_grid_address[i] = _get_relative_grid_addresses_from_main_diagonal(
                i
            )[0]

    return relative_grid_address


def get_tetrahedra_integration_weight(omegas, tetrahedra_omegas, function="I"):
    """Return integration weights.

    Parameters
    ----------
    omegas : float or list of float values
        Energy(s) at which the integration weight(s) are computed.
    tetrahedra_omegas : ndarray of list of list
        Energies at vertices of 24 tetrahedra
        shape=(24, 4)
        dytpe='double'
    function : str, 'I' or 'J'
        'J' is for intetration and 'I' is for its derivative.

    """
    try:
        import phonopy._phonopy as phonoc
    except ImportError:
        import sys

        print("Phonopy C-extension has to be built properly.")
        sys.exit(1)

    if isinstance(omegas, float):
        return phonoc.tetrahedra_integration_weight(
            omegas, np.array(tetrahedra_omegas, dtype="double", order="C"), function
        )
    else:
        integration_weights = np.zeros(len(omegas), dtype="double")
        phonoc.tetrahedra_integration_weight_at_omegas(
            integration_weights,
            np.array(omegas, dtype="double"),
            np.array(tetrahedra_omegas, dtype="double", order="C"),
            function,
        )
        return integration_weights


class TetrahedronMethod:
    """Class to perform linear tetrahedron method on regular grid locally."""

    def __init__(self, primitive_vectors, mesh=None, lang="C"):
        """Init method.

        Parameters
        ----------
        primitive_vectors : ndarray
            a, b, c in column vectors in reciprocal space. This is only used to
            determine a main diagonal of parallelepiped. If None, the first
            main diagonal in the dataset is chosen.
            shape=(3, 3)
        mesh : array_like
            Mesh numbers.

        """
        if mesh is None:
            mesh = [1, 1, 1]
        if primitive_vectors is None:
            self._primitive_vectors = None
        else:
            self._primitive_vectors = (
                np.array(primitive_vectors, dtype="double", order="C") / mesh
            )
        self._lang = lang
        self._vertices = None
        self._relative_grid_addresses = None
        self._central_indices = None
        self._tetrahedra_omegas = None
        self._sort_indices = None
        self._omegas = None
        self._integration_weight = None
        self._set_relative_grid_addresses(lang=self._lang)

    def run(self, omegas, value="I"):
        """Perform tetrahedron method.

        Parameters
        ----------
        value : str of "I" or "J"
            "I": Imaginary part (delta function). "J": Integral function of
            imaginary part.

        Note
        ----
        "C" or "Py" here has to be consistent with that of
        self._set_relative_grid_addresses().

        """
        if self._lang == "C":
            self._run_c(omegas, value=value)
        else:
            self._run_py(omegas, value=value)

    @property
    def tetrahedra(self):
        """Return relative grid addresses at vertices of tetrahedra."""
        return self._relative_grid_addresses

    def get_tetrahedra(self):
        """Return relative grid addresses at vertices of tetrahedra."""
        warnings.warn(
            (
                "TetrahedronMethod.get_tetrahedra() is deprecated. "
                "Use TetrahedronMethod.tetrahedra attribute instead."
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return self.tetrahedra

    def get_unique_tetrahedra_vertices(self):
        """Return unique grid indices in neighboring tetrahedra."""
        unique_vertices = []
        for adrs in self._relative_grid_addresses.reshape(-1, 3):
            found = False
            for uadrs in unique_vertices:
                if (uadrs == adrs).all():
                    found = True
                    break
            if not found:
                unique_vertices.append(adrs)
        return np.array(unique_vertices, dtype="int64", order="C")

    def set_tetrahedra_omegas(self, tetrahedra_omegas):
        """Set values on vertices of tetrahedra.

        tetrahedra_omegas: (24, 4) omegas at self._relative_grid_addresses

        """
        self._tetrahedra_omegas = tetrahedra_omegas

    def get_integration_weight(self):
        """Return integration weights."""
        return self._integration_weight

    def _set_relative_grid_addresses(self, lang="C"):
        """Set dataset of relative grid addresses."""
        if lang == "C":
            if self._primitive_vectors is None:
                rga = np.array(
                    get_all_tetrahedra_relative_grid_address()[0],
                    dtype="int64",
                    order="C",
                )
            else:
                rga = get_tetrahedra_relative_grid_address(self._primitive_vectors)
            self._relative_grid_addresses = rga
        else:
            if self._primitive_vectors is None:
                (
                    relative_grid_addresses,
                    central_indices,
                ) = _get_relative_grid_addresses_from_main_diagonal(0)
            else:
                (
                    relative_grid_addresses,
                    central_indices,
                ) = _get_relative_grid_addresses_from_microzone_lattice(
                    self._primitive_vectors
                )
            self._relative_grid_addresses = relative_grid_addresses
            self._central_indices = central_indices

    def _run_c(self, omegas, value="I"):
        self._integration_weight = get_tetrahedra_integration_weight(
            omegas, self._tetrahedra_omegas, function=value
        )

    def _run_py(self, omegas, value="I"):
        if isinstance(omegas, float) or isinstance(omegas, int):
            iw = self._get_integration_weight_py(omegas, value=value)
        else:
            iw = np.zeros(len(omegas), dtype="double")
            for i, omega in enumerate(omegas):
                iw[i] = self._get_integration_weight_py(omega, value=value)
        self._integration_weight = iw

    def _get_integration_weight_py(self, omega, value="I"):
        if value == "I":
            IJ = self._I
            gn = self._g
        else:
            IJ = self._J
            gn = self._n

        self._sort_indices = np.argsort(self._tetrahedra_omegas, axis=1)
        sum_value = 0.0
        self._omega = omega
        for omegas, indices, ci in zip(
            self._tetrahedra_omegas, self._sort_indices, self._central_indices
        ):
            self._vertices_omegas = omegas[indices]
            # i_where = np.where(omega < self._vertices_omegas)[0]
            # if len(i_where):
            #     i = i_where[0]
            # else:
            #     i = 4
            v = self._vertices_omegas
            if omega < v[0]:
                sum_value += IJ(0, np.where(indices == ci)[0][0]) * gn(0)
            elif v[0] < omega and omega < v[1]:
                sum_value += IJ(1, np.where(indices == ci)[0][0]) * gn(1)
            elif v[1] < omega and omega < v[2]:
                sum_value += IJ(2, np.where(indices == ci)[0][0]) * gn(2)
            elif v[2] < omega and omega < v[3]:
                sum_value += IJ(3, np.where(indices == ci)[0][0]) * gn(3)
            elif v[3] < omega:
                sum_value += IJ(4, np.where(indices == ci)[0][0]) * gn(4)

        return sum_value / 6

    def _f(self, n, m):
        return (self._omega - self._vertices_omegas[m]) / (
            self._vertices_omegas[n] - self._vertices_omegas[m]
        )

    def _J(self, i, ci):
        if i == 0:
            return self._J_0()
        elif i == 1:
            if ci == 0:
                return self._J_10()
            elif ci == 1:
                return self._J_11()
            elif ci == 2:
                return self._J_12()
            elif ci == 3:
                return self._J_13()
            else:
                raise RuntimeError("Unexpected condition encountered.")
        elif i == 2:
            if ci == 0:
                return self._J_20()
            elif ci == 1:
                return self._J_21()
            elif ci == 2:
                return self._J_22()
            elif ci == 3:
                return self._J_23()
            else:
                raise RuntimeError("Unexpected condition encountered.")
        elif i == 3:
            if ci == 0:
                return self._J_30()
            elif ci == 1:
                return self._J_31()
            elif ci == 2:
                return self._J_32()
            elif ci == 3:
                return self._J_33()
            else:
                raise RuntimeError("Unexpected condition encountered.")
        elif i == 4:
            return self._J_4()
        else:
            raise RuntimeError("Unexpected condition encountered.")

    def _I(self, i, ci):
        if i == 0:
            return self._I_0()
        elif i == 1:
            if ci == 0:
                return self._I_10()
            elif ci == 1:
                return self._I_11()
            elif ci == 2:
                return self._I_12()
            elif ci == 3:
                return self._I_13()
            else:
                raise RuntimeError("Unexpected condition encountered.")
        elif i == 2:
            if ci == 0:
                return self._I_20()
            elif ci == 1:
                return self._I_21()
            elif ci == 2:
                return self._I_22()
            elif ci == 3:
                return self._I_23()
            else:
                raise RuntimeError("Unexpected condition encountered.")
        elif i == 3:
            if ci == 0:
                return self._I_30()
            elif ci == 1:
                return self._I_31()
            elif ci == 2:
                return self._I_32()
            elif ci == 3:
                return self._I_33()
            else:
                raise RuntimeError("Unexpected condition encountered.")
        elif i == 4:
            return self._I_4()
        else:
            raise RuntimeError("Unexpected condition encountered.")

    def _n(self, i):
        if i == 0:
            return self._n_0()
        elif i == 1:
            return self._n_1()
        elif i == 2:
            return self._n_2()
        elif i == 3:
            return self._n_3()
        elif i == 4:
            return self._n_4()
        else:
            raise RuntimeError("Unexpected condition encountered.")

    def _g(self, i):
        if i == 0:
            return self._g_0()
        elif i == 1:
            return self._g_1()
        elif i == 2:
            return self._g_2()
        elif i == 3:
            return self._g_3()
        elif i == 4:
            return self._g_4()
        else:
            raise RuntimeError("Unexpected condition encountered.")

    def _n_0(self):
        """n0.

        omega < omega1

        """
        return 0.0

    def _n_1(self):
        """n1.

        omega1 < omega < omega2

        """
        return self._f(1, 0) * self._f(2, 0) * self._f(3, 0)

    def _n_2(self):
        """n2.

        omega2 < omega < omega3

        """
        return (
            self._f(3, 1) * self._f(2, 1)
            + self._f(3, 0) * self._f(1, 3) * self._f(2, 1)
            + self._f(3, 0) * self._f(2, 0) * self._f(1, 2)
        )

    def _n_3(self):
        """n3.

        omega2 < omega < omega3

        """
        return 1.0 - self._f(0, 3) * self._f(1, 3) * self._f(2, 3)

    def _n_4(self):
        """n4.

        omega4 < omega

        """
        return 1.0

    def _g_0(self):
        """g0.

        omega < omega1

        """
        return 0.0

    def _g_1(self):
        """g1.

        omega1 < omega < omega2

        """
        # return 3 * self._n_1() / (self._omega - self._vertices_omegas[0])
        return (
            3
            * self._f(1, 0)
            * self._f(2, 0)
            / (self._vertices_omegas[3] - self._vertices_omegas[0])
        )

    def _g_2(self):
        """g2.

        omega2 < omega < omega3

        """
        return (
            3
            / (self._vertices_omegas[3] - self._vertices_omegas[0])
            * (self._f(1, 2) * self._f(2, 0) + self._f(2, 1) * self._f(1, 3))
        )

    def _g_3(self):
        """g3.

        omega3 < omega < omega4

        """
        # return 3 * (1.0 - self._n_3()) / (self._vertices_omegas[3] - self._omega)
        return (
            3
            * self._f(1, 3)
            * self._f(2, 3)
            / (self._vertices_omegas[3] - self._vertices_omegas[0])
        )

    def _g_4(self):
        """g4.

        omega4 < omega

        """
        return 0.0

    def _J_0(self):
        return 0.0

    def _J_10(self):
        return (1.0 + self._f(0, 1) + self._f(0, 2) + self._f(0, 3)) / 4

    def _J_11(self):
        return self._f(1, 0) / 4

    def _J_12(self):
        return self._f(2, 0) / 4

    def _J_13(self):
        return self._f(3, 0) / 4

    def _J_20(self):
        return (
            (
                self._f(3, 1) * self._f(2, 1)
                + self._f(3, 0) * self._f(1, 3) * self._f(2, 1) * (1.0 + self._f(0, 3))
                + self._f(3, 0)
                * self._f(2, 0)
                * self._f(1, 2)
                * (1.0 + self._f(0, 3) + self._f(0, 2))
            )
            / 4
            / self._n_2()
        )

    def _J_21(self):
        return (
            (
                self._f(3, 1) * self._f(2, 1) * (1.0 + self._f(1, 3) + self._f(1, 2))
                + self._f(3, 0)
                * self._f(1, 3)
                * self._f(2, 1)
                * (self._f(1, 3) + self._f(1, 2))
                + self._f(3, 0) * self._f(2, 0) * self._f(1, 2) * self._f(1, 2)
            )
            / 4
            / self._n_2()
        )

    def _J_22(self):
        return (
            (
                self._f(3, 1) * self._f(2, 1) * self._f(2, 1)
                + self._f(3, 0) * self._f(1, 3) * self._f(2, 1) * self._f(2, 1)
                + self._f(3, 0)
                * self._f(2, 0)
                * self._f(1, 2)
                * (self._f(2, 1) + self._f(2, 0))
            )
            / 4
            / self._n_2()
        )

    def _J_23(self):
        return (
            (
                self._f(3, 1) * self._f(2, 1) * self._f(3, 1)
                + self._f(3, 0)
                * self._f(1, 3)
                * self._f(2, 1)
                * (self._f(3, 1) + self._f(3, 0))
                + self._f(3, 0) * self._f(2, 0) * self._f(1, 2) * self._f(3, 0)
            )
            / 4
            / self._n_2()
        )

    def _J_30(self):
        return (
            (1.0 - self._f(0, 3) ** 2 * self._f(1, 3) * self._f(2, 3)) / 4 / self._n_3()
        )

    def _J_31(self):
        return (
            (1.0 - self._f(0, 3) * self._f(1, 3) ** 2 * self._f(2, 3)) / 4 / self._n_3()
        )

    def _J_32(self):
        return (
            (1.0 - self._f(0, 3) * self._f(1, 3) * self._f(2, 3) ** 2) / 4 / self._n_3()
        )

    def _J_33(self):
        return (
            (
                1.0
                - self._f(0, 3)
                * self._f(1, 3)
                * self._f(2, 3)
                * (1.0 + self._f(3, 0) + self._f(3, 1) + self._f(3, 2))
            )
            / 4
            / self._n_3()
        )

    def _J_4(self):
        return 0.25

    def _I_0(self):
        return 0.0

    def _I_10(self):
        return (self._f(0, 1) + self._f(0, 2) + self._f(0, 3)) / 3

    def _I_11(self):
        return self._f(1, 0) / 3

    def _I_12(self):
        return self._f(2, 0) / 3

    def _I_13(self):
        return self._f(3, 0) / 3

    def _I_20(self):
        return (
            self._f(0, 3)
            + self._f(0, 2)
            * self._f(2, 0)
            * self._f(1, 2)
            / (self._f(1, 2) * self._f(2, 0) + self._f(2, 1) * self._f(1, 3))
        ) / 3

    def _I_21(self):
        return (
            self._f(1, 2)
            + self._f(1, 3) ** 2
            * self._f(2, 1)
            / (self._f(1, 2) * self._f(2, 0) + self._f(2, 1) * self._f(1, 3))
        ) / 3

    def _I_22(self):
        return (
            self._f(2, 1)
            + self._f(2, 0) ** 2
            * self._f(1, 2)
            / (self._f(1, 2) * self._f(2, 0) + self._f(2, 1) * self._f(1, 3))
        ) / 3

    def _I_23(self):
        return (
            self._f(3, 0)
            + self._f(3, 1)
            * self._f(1, 3)
            * self._f(2, 1)
            / (self._f(1, 2) * self._f(2, 0) + self._f(2, 1) * self._f(1, 3))
        ) / 3

    def _I_30(self):
        return self._f(0, 3) / 3

    def _I_31(self):
        return self._f(1, 3) / 3

    def _I_32(self):
        return self._f(2, 3) / 3

    def _I_33(self):
        return (self._f(3, 0) + self._f(3, 1) + self._f(3, 2)) / 3

    def _I_4(self):
        return 0.0


def _create_tetrahedra(shortest_main_diagonal):
    """Create dataset of six vertices.

        6-------7
        /|      /|
        / |     / |
        4-------5  |
        |  2----|--3
        | /     | /
        |/      |/
        0-------1

    i: vec        neighbours
    0: O          1, 2, 4
    1: a          0, 3, 5
    2: b          0, 3, 6
    3: a + b      1, 2, 7
    4: c          0, 5, 6
    5: c + a      1, 4, 7
    6: c + b      2, 4, 7
    7: c + a + b  3, 5, 6

    Four main diagonals: 0-7, 1-6, 2-5, 3-4

    """
    if shortest_main_diagonal == 0:
        pairs = ((1, 3), (1, 5), (2, 3), (2, 6), (4, 5), (4, 6))
        six_tetras = np.sort([[0, 7] + list(x) for x in pairs])
    elif shortest_main_diagonal == 1:
        pairs = ((0, 2), (0, 4), (2, 3), (3, 7), (4, 5), (5, 7))
        six_tetras = np.sort([[1, 6] + list(x) for x in pairs])
    elif shortest_main_diagonal == 2:
        pairs = ((0, 1), (0, 4), (1, 3), (3, 7), (4, 6), (6, 7))
        six_tetras = np.sort([[2, 5] + list(x) for x in pairs])
    elif shortest_main_diagonal == 3:
        pairs = ((0, 1), (0, 2), (1, 5), (2, 6), (5, 7), (6, 7))
        six_tetras = np.sort([[3, 4] + list(x) for x in pairs])
    else:
        RuntimeError("Unexpected condition encountered.")

    return six_tetras


def _get_relative_grid_addresses_from_six_tetrahedra(six_tetras):
    parallelepiped_vertices = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
            [1, 1, 1],
        ],
        dtype="int64",
        order="C",
    )
    relative_grid_addresses = np.zeros((24, 4, 3), dtype="int64", order="C")
    central_indices = np.zeros(24, dtype="int64")
    pos = 0
    for i in range(8):
        ppd_shifted = parallelepiped_vertices - parallelepiped_vertices[i]
        for tetra in six_tetras:
            if i in tetra:
                central_indices[pos] = np.where(tetra == i)[0][0]
                relative_grid_addresses[pos, :, :] = ppd_shifted[tetra]
                pos += 1
    return relative_grid_addresses, central_indices


def _get_relative_grid_addresses_from_microzone_lattice(microzone_lattice):
    """Return relative grid addresses of given microzone lattice.

    microzone lattice is given by column vectors.

    """
    a, b, c = microzone_lattice.T
    main_diagonals = np.array([a + b + c, -a + b + c, a - b + c, a + b - c])
    shortest_main_diagonal = np.argmin(np.sum(main_diagonals**2, axis=1))
    return _get_relative_grid_addresses_from_main_diagonal(shortest_main_diagonal)


def _get_relative_grid_addresses_from_main_diagonal(main_diagonal):
    """Return relative grid addresses of given main diagonal."""
    six_tetras = _create_tetrahedra(main_diagonal)
    return _get_relative_grid_addresses_from_six_tetrahedra(six_tetras)



"""Tests for symmetry tools."""

import numpy as np
import pytest

from phonopy import Phonopy
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import get_supercell
from phonopy.structure.symmetry import (
    Symmetry,
    _get_mapping_between_cells,
    collect_unique_rotations,
    symmetrize_borns_and_epsilon,
)


def test_get_map_operations(nacl_unitcell_order1):
    """Test get_map_operations()."""
    symprec = 1e-5
    cell = nacl_unitcell_order1
    scell = get_supercell(cell, np.diag([2, 2, 2]), symprec=symprec)
    symmetry = Symmetry(scell, symprec=symprec)
    map_ops = symmetry.get_map_operations().copy()
    # start = time.time()
    # symmetry._set_map_operations()
    # end = time.time()
    # print(end - start)
    # map_ops_old = symmetry.get_map_operations().copy()
    # assert (map_ops == map_ops_old).all()
    map_atoms = symmetry.get_map_atoms()
    positions = scell.scaled_positions
    rotations = symmetry.symmetry_operations["rotations"]
    translations = symmetry.symmetry_operations["translations"]
    for i, (op_i, atom_i) in enumerate(zip(map_ops, map_atoms)):
        r_pos = np.dot(rotations[op_i], positions[i]) + translations[op_i]
        diff = positions[atom_i] - r_pos
        diff -= np.rint(diff)
        assert (diff < symprec).all()


def test_collinear_magnetic_symmetry(convcell_cr: PhonopyAtoms):
    """Test symmetry search with collinear magnetic moments."""
    symprec = 1e-5
    symmetry_nonspin = Symmetry(convcell_cr, symprec=symprec)
    atom_map_nonspin = symmetry_nonspin.get_map_atoms()
    len_sym_nonspin = len(symmetry_nonspin.symmetry_operations["rotations"])

    cell_withspin = convcell_cr.copy()
    cell_withspin.magnetic_moments = [1, -1]
    symmetry_withspin = Symmetry(cell_withspin, symprec=symprec)
    atom_map_withspin = symmetry_withspin.get_map_atoms()
    len_sym_withspin = len(symmetry_withspin.symmetry_operations["rotations"])

    cell_brokenspin = convcell_cr.copy()
    cell_brokenspin.magnetic_moments = [1, -2]
    symmetry_brokenspin = Symmetry(cell_brokenspin, symprec=symprec)
    atom_map_brokenspin = symmetry_brokenspin.get_map_atoms()
    len_sym_brokenspin = len(symmetry_brokenspin.symmetry_operations["rotations"])

    assert (atom_map_nonspin == atom_map_withspin).all()
    assert (atom_map_nonspin != atom_map_brokenspin).any()
    assert len_sym_nonspin == len_sym_withspin
    assert len_sym_nonspin != len_sym_brokenspin


@pytest.mark.parametrize("is_flat", [False, True])
def test_non_collinear_magnetic_symmetry(convcell_cr: PhonopyAtoms, is_flat: bool):
    """Test symmetry search with non-collinear magnetic moments."""
    symprec = 1e-5
    symmetry_nonspin = Symmetry(convcell_cr, symprec=symprec)
    atom_map_nonspin = symmetry_nonspin.get_map_atoms()
    len_sym_nonspin = len(symmetry_nonspin.symmetry_operations["rotations"])

    cell_withspin = convcell_cr.copy()
    if is_flat:
        cell_withspin.magnetic_moments = [0, 0, 1, 0, 0, -1]
    else:
        cell_withspin.magnetic_moments = [[0, 0, 1], [0, 0, -1]]
    symmetry_withspin = Symmetry(cell_withspin, symprec=symprec)
    atom_map_withspin = symmetry_withspin.get_map_atoms()
    len_sym_withspin = len(symmetry_withspin.symmetry_operations["rotations"])

    cell_brokenspin = convcell_cr.copy()
    if is_flat:
        cell_brokenspin.magnetic_moments = [1, 1, 1, -2, -2, -2]
    else:
        cell_brokenspin.magnetic_moments = [[1, 1, 1], [-2, -2, -2]]
    symmetry_brokenspin = Symmetry(cell_brokenspin, symprec=symprec)
    atom_map_brokenspin = symmetry_brokenspin.get_map_atoms()
    len_sym_brokenspin = len(symmetry_brokenspin.symmetry_operations["rotations"])

    assert (atom_map_nonspin == atom_map_withspin).all()
    assert (atom_map_nonspin == atom_map_brokenspin).any()
    assert len_sym_nonspin == 96
    assert len_sym_withspin == 32
    assert len_sym_brokenspin == 12


def test_symmetrize_borns_and_epsilon_nacl(ph_nacl: Phonopy):
    """Test symmetrization of Born charges and dielectric tensors by NaCl."""
    nac_params = ph_nacl.nac_params
    borns, epsilon = symmetrize_borns_and_epsilon(
        nac_params["born"], nac_params["dielectric"], ph_nacl.primitive
    )
    np.testing.assert_allclose(borns, nac_params["born"], atol=1e-8)
    np.testing.assert_allclose(epsilon, nac_params["dielectric"], atol=1e-8)


def test_symmetrize_borns_and_epsilon_tio2(ph_tio2: Phonopy):
    """Test symmetrization of Born charges and dielectric tensors by TiO2."""
    nac_params = ph_tio2.nac_params
    borns, epsilon = symmetrize_borns_and_epsilon(
        nac_params["born"], nac_params["dielectric"], ph_tio2.primitive
    )
    # np.testing.assert_allclose(borns, nac_params['born'], atol=1e-8)
    np.testing.assert_allclose(epsilon, nac_params["dielectric"], atol=1e-8)


def test_Symmetry_pointgroup(ph_tio2: Phonopy):
    """Test for point group symbol."""
    assert ph_tio2.symmetry.pointgroup_symbol == r"4/mmm"


def test_Symmetry_nosym_s2p_map(nacl_unitcell_order1: PhonopyAtoms):
    """Test Symmetry with is_symmetry=False and s2p_map.

    This situation happens when making Symmetry with nosym for supercell in
    the Phonopy class.

    """
    ph = Phonopy(
        nacl_unitcell_order1,
        supercell_matrix=[2, 2, 2],
        primitive_matrix="F",
        is_symmetry=False,
    )
    # for i, v in enumerate(ph.symmetry.symmetry_operations["translations"]):
    #     print("[", ", ".join(f"{x}" for x in v), "],")
    np.testing.assert_equal(
        ph.symmetry.symmetry_operations["translations"],
        [
            [0.0, 0.0, 0.0],
            [0.5, 0.0, 0.0],
            [0.0, 0.5, 0.0],
            [0.5, 0.5, 0.0],
            [0.0, 0.0, 0.5],
            [0.5, 0.0, 0.5],
            [0.0, 0.5, 0.5],
            [0.5, 0.5, 0.5],
            [0.0, 0.25, 0.25],
            [0.5, 0.25, 0.25],
            [0.0, 0.75, 0.25],
            [0.5, 0.75, 0.25],
            [0.0, 0.25, 0.75],
            [0.5, 0.25, 0.75],
            [0.0, 0.75, 0.75],
            [0.5, 0.75, 0.75],
            [0.25, 0.0, 0.25],
            [0.75, 0.0, 0.25],
            [0.25, 0.5, 0.25],
            [0.75, 0.5, 0.25],
            [0.25, 0.0, 0.75],
            [0.75, 0.0, 0.75],
            [0.25, 0.5, 0.75],
            [0.75, 0.5, 0.75],
            [0.25, 0.25, 0.0],
            [0.75, 0.25, 0.0],
            [0.25, 0.75, 0.0],
            [0.75, 0.75, 0.0],
            [0.25, 0.25, 0.5],
            [0.75, 0.25, 0.5],
            [0.25, 0.75, 0.5],
            [0.75, 0.75, 0.5],
        ],
    )


def test_with_pmat_and_smat(ph_nacl: Phonopy):
    """Test Born charges and dielectric tensor symmetrization with pmat and smat."""
    pcell = ph_nacl.primitive
    scell = ph_nacl.supercell
    idx = [scell.u2u_map[i] for i in scell.s2u_map[pcell.p2s_map]]
    uborns, uepsilon = _get_nac_params_in_unitcell(ph_nacl)
    borns, epsilon = symmetrize_borns_and_epsilon(
        uborns,
        uepsilon,
        ph_nacl.unitcell,
        primitive_matrix=ph_nacl.primitive_matrix,
        supercell_matrix=ph_nacl.supercell_matrix,
    )
    np.testing.assert_allclose(borns, uborns[idx], atol=1e-8)
    np.testing.assert_allclose(epsilon, uepsilon, atol=1e-8)


def test_with_pcell(ph_nacl: Phonopy):
    """Test Born charges and dielectric tensor symmetrization with pcell."""
    pcell = ph_nacl.primitive
    scell = ph_nacl.supercell
    idx = [scell.u2u_map[i] for i in scell.s2u_map[pcell.p2s_map]]
    idx2 = _get_mapping_between_cells(pcell, pcell)
    np.testing.assert_array_equal(idx2, np.arange(len(pcell)))

    uborns, uepsilon = _get_nac_params_in_unitcell(ph_nacl)
    borns, epsilon = symmetrize_borns_and_epsilon(
        uborns, uepsilon, ph_nacl.unitcell, primitive=pcell
    )
    np.testing.assert_allclose(borns, uborns[idx][idx2], atol=1e-8)
    np.testing.assert_allclose(epsilon, uepsilon, atol=1e-8)


def test_site_symmetry(ph_sno2: Phonopy):
    """Test site symmetry operations."""
    site_sym0 = ph_sno2.symmetry.get_site_symmetry(0)
    ref0 = [
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        -1,
        0,
        -1,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        -1,
        0,
        -1,
        0,
        -1,
        0,
        0,
        0,
        0,
        -1,
    ]
    site_sym36 = ph_sno2.symmetry.get_site_symmetry(36)
    ref36 = [
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        -1,
        0,
        1,
        0,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        1,
        0,
        0,
        0,
        0,
        -1,
    ]
    np.testing.assert_array_equal(site_sym0.ravel(), ref0)
    np.testing.assert_array_equal(site_sym36.ravel(), ref36)


def test_collect_unique_rotations(ph_nacl: Phonopy):
    """Test collect_unique_rotations function."""
    rotations = ph_nacl.symmetry.symmetry_operations["rotations"]
    ptg = collect_unique_rotations(rotations)
    assert len(rotations) == 1536
    assert len(ptg) == 48
    assert len(ptg) == len(ph_nacl.symmetry.pointgroup_operations)


def test_reciprocal_operations(ph_zr3n4: Phonopy):
    """Test reciprocal operations.

    Zr3N4 is a non-centrosymmetric crystal.

    """
    ptg = ph_zr3n4.symmetry.pointgroup_operations
    rops = ph_zr3n4.symmetry.reciprocal_operations
    matches = []
    for r in ptg:
        for i, rec_r in enumerate(rops):
            if (r == rec_r).all():
                matches.append(i)
                break
    assert len(np.unique(matches)) == len(ptg)
    found_inv = False
    for rec_r in rops:
        if (rec_r == -np.eye(3, dtype=int)).all():
            found_inv = True
            break
    assert found_inv


def _get_nac_params_in_unitcell(ph: Phonopy):
    nac_params = ph.nac_params
    uepsilon = nac_params["dielectric"]
    pborns = nac_params["born"]
    s2p_map = ph.primitive.s2p_map
    p2p_map = ph.primitive.p2p_map
    s2pp_map = [p2p_map[i] for i in s2p_map]
    sborns = pborns[s2pp_map]
    uborns = np.array([sborns[i] for i in ph.supercell.u2s_map])

    return uborns, uepsilon



"""Tests for dynmat_to_fc, inverse phonon transformation."""

import os
from io import StringIO

import numpy as np
import pytest

from phonopy import Phonopy
from phonopy.harmonic.dynmat_to_fc import (
    DynmatToForceConstants,
    get_commensurate_points,
    get_commensurate_points_in_integers,
    ph2ph,
)
from phonopy.physical_units import get_physical_units

data_dir = os.path.dirname(os.path.abspath(__file__))


comm_points_str = """1  0.00  0.00  0.00
2  0.00  0.25  0.25
3  0.00  0.50  0.50
4  0.00  0.75  0.75
5  0.25  0.00  0.25
6  0.25  0.25  0.50
7  0.25  0.50  0.75
8  0.25  0.75  0.00
9  0.50  0.00  0.50
10  0.50  0.25  0.75
11  0.50  0.50  0.00
12  0.50  0.75  0.25
13  0.75  0.00  0.75
14  0.75  0.25  0.00
15  0.75  0.50  0.25
16  0.75  0.75  0.50
17  0.25  0.25  0.00
18  0.25  0.50  0.25
19  0.25  0.75  0.50
20  0.25  0.00  0.75
21  0.50  0.25  0.25
22  0.50  0.50  0.50
23  0.50  0.75  0.75
24  0.50  0.00  0.00
25  0.75  0.25  0.50
26  0.75  0.50  0.75
27  0.75  0.75  0.00
28  0.75  0.00  0.25
29  0.00  0.25  0.75
30  0.00  0.50  0.00
31  0.00  0.75  0.25
32  0.00  0.00  0.50"""


def test_get_commensurate_points():
    """Test for getting commensurate points."""
    comm_points, _ = _get_commensurate_points()
    _compare(comm_points)


def test_get_commensurate_points_in_integers():
    """Test for getting commensurate points represented by integers."""
    comm_points_ref, smat = _get_commensurate_points()
    comm_points = get_commensurate_points_in_integers(smat)
    comm_points = comm_points / np.linalg.det(smat)

    all_indices = []
    for cpt in comm_points:
        diff = comm_points_ref - cpt
        diff -= np.rint(diff)
        dist2 = (diff**2).sum(axis=1)
        indices = np.where(dist2 < 1e-5)[0]
        assert len(indices) == 1
        all_indices.append(indices[0])

    all_indices.sort()
    np.testing.assert_array_equal(all_indices, np.arange(len(comm_points)))


def _get_commensurate_points():
    smat = np.dot([[-1, 1, 1], [1, -1, 1], [1, 1, -1]], np.diag([2, 2, 2]))
    return get_commensurate_points(smat), smat


def _compare(comm_points):
    comm_points_ref = np.loadtxt(StringIO(comm_points_str), dtype="double")
    diff = comm_points_ref[:, 1:] - comm_points
    np.testing.assert_allclose(diff, np.rint(diff), atol=1e-3)


def _write(comm_points, filename="comm_points.dat"):
    with open(os.path.join(data_dir, filename), "w") as w:
        lines = []
        for i, p in enumerate(comm_points):
            lines.append("%d %5.2f %5.2f %5.2f" % ((i + 1,) + tuple(p)))
        w.write("\n".join(lines))


def test_ph2ph_with_nac(ph_nacl):
    """Test transformation of phonon to force constants with NAC."""
    ph = ph_nacl
    fc333 = ph2ph(ph, np.diag([3, 3, 3]), with_nac=True).force_constants
    _phonons_allclose(ph, fc333)


def test_ph2ph_without_nac(ph_nacl_nonac):
    """Test transformation of phonon to force constants without NAC."""
    ph = ph_nacl_nonac
    fc333 = ph2ph(ph, np.diag([3, 3, 3]), with_nac=False).force_constants
    _phonons_allclose(ph, fc333)


def _phonons_allclose(ph: Phonopy, fc333):
    ph333 = Phonopy(
        ph.unitcell, supercell_matrix=[3, 3, 3], primitive_matrix=ph.primitive_matrix
    )
    ph333.force_constants = fc333
    ph333.nac_params = ph.nac_params
    comm_points = _get_comm_points(ph)
    ph.run_qpoints(comm_points)
    ph333.run_qpoints(comm_points)
    np.testing.assert_allclose(
        ph.get_qpoints_dict()["frequencies"],
        ph333.get_qpoints_dict()["frequencies"],
        atol=1e-5,
    )


def _get_comm_points(ph):
    smat = np.dot(np.linalg.inv(ph.primitive.primitive_matrix), ph.supercell_matrix)
    smat = np.rint(smat).astype(int)
    comm_points = get_commensurate_points(smat)
    return comm_points


def test_with_eigenvalues(ph_nacl, ph_nacl_nonac):
    """Test transformation from eigensolutions to dynamical matrix."""
    for ph in (ph_nacl_nonac, ph_nacl):
        d2f = DynmatToForceConstants(ph.primitive, ph.supercell)
        ph.run_qpoints(
            d2f.commensurate_points,
            with_eigenvectors=True,
            with_dynamical_matrices=True,
        )
        ph_dict = ph.get_qpoints_dict()
        eigenvalues = (
            ph_dict["frequencies"] / get_physical_units().DefaultToTHz
        ) ** 2 * np.sign(ph_dict["frequencies"])
        d2f.create_dynamical_matrices(
            eigenvalues=eigenvalues, eigenvectors=ph_dict["eigenvectors"]
        )
        d2f.run()
        np.testing.assert_allclose(ph.force_constants, d2f.force_constants, atol=1e-5)


@pytest.mark.parametrize(
    "is_nac,lang", [(False, "C"), (True, "C"), (False, "Py"), (True, "Py")]
)
def test_with_dynamical_matrices(ph_nacl, ph_nacl_nonac, is_nac, lang):
    """Test transformation from dynamical matrix to force constants."""
    if is_nac:
        ph = ph_nacl
    else:
        ph = ph_nacl_nonac

    d2f = DynmatToForceConstants(ph.primitive, ph.supercell)
    ph.run_qpoints(d2f.commensurate_points, with_dynamical_matrices=True)
    ph_dict = ph.get_qpoints_dict()
    d2f.dynamical_matrices = ph_dict["dynamical_matrices"]
    d2f.run(lang=lang)
    np.testing.assert_allclose(ph.force_constants, d2f.force_constants, atol=1e-5)



"""A script to generate supercells with displacements for LAMMPS."""

import phonopy
from phonopy.interface.calculator import write_supercells_with_displacements
from phonopy.interface.phonopy_yaml import read_cell_yaml

cell = read_cell_yaml("phonopy_unitcell.yaml")
ph = phonopy.load(
    unitcell=cell,
    primitive_matrix="auto",
    supercell_matrix=[2, 2, 2],
    calculator="lammps",
)
ph.generate_displacements()
ph.save("phonopy_disp.yaml")
write_supercells_with_displacements(
    "lammps", ph.supercell, ph.supercells_with_displacements
)



"""Routines to handle displacements in supercells."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import Optional

import numpy as np

directions_axis = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

directions_diag = np.array(
    [
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
        [1, 1, 0],
        [1, 0, 1],
        [0, 1, 1],
        [1, -1, 0],
        [1, 0, -1],
        [0, 1, -1],
        [1, 1, 1],
        [1, 1, -1],
        [1, -1, 1],
        [-1, 1, 1],
    ]
)


def directions_to_displacement_dataset(displacement_directions, distance, supercell):
    """Transform displacement directions to displacements in Cartesian coordinates."""
    lattice = supercell.cell
    first_atoms = []
    for disp in displacement_directions:
        direction = disp[1:]
        disp_cartesian = np.dot(direction, lattice)
        disp_cartesian *= distance / np.linalg.norm(disp_cartesian)
        first_atoms.append(
            {"number": int(disp[0]), "displacement": disp_cartesian.tolist()}
        )
    displacement_dataset = {
        "natom": len(supercell),
        "first_atoms": first_atoms,
    }

    return displacement_dataset


def get_least_displacements(
    symmetry,
    is_plusminus="auto",
    is_diagonal=True,
    is_trigonal=False,
    log_level=0,
):
    """Return a set of displacements.

    Returns
    -------
    array_like
        List of directions with respect to axes. This gives only the
        symmetrically non equivalent directions. The format is like:
           [[0, 1, 0, 0],
            [7, 1, 0, 1], ...]
        where each list is defined by:
           First value:      Atom index in supercell starting with 0
           Second to fourth: If the direction is displaced or not (1, 0, or -1)
                             with respect to the axes.

    """
    displacements = []
    if is_diagonal:
        directions = directions_diag
    else:
        directions = directions_axis

    if log_level > 2:
        print("Site point symmetry:")

    for atom_num in symmetry.get_independent_atoms():
        site_symmetry = symmetry.get_site_symmetry(atom_num)

        if log_level > 2:
            print("Atom %d" % (atom_num + 1))
            for i, rot in enumerate(site_symmetry):
                print("----%d----" % (i + 1))
                for v in rot:
                    print("%2d %2d %2d" % tuple(v))

        for disp in get_displacement(site_symmetry, directions, is_trigonal, log_level):
            displacements.append([atom_num, disp[0], disp[1], disp[2]])
            if is_plusminus == "auto":
                if is_minus_displacement(disp, site_symmetry):
                    displacements.append([atom_num, -disp[0], -disp[1], -disp[2]])
            elif is_plusminus is True:
                displacements.append([atom_num, -disp[0], -disp[1], -disp[2]])

    return displacements


def get_displacement(
    site_symmetry, directions=directions_diag, is_trigonal=False, log_level=0
):
    """Return displacement directions for one atom."""
    # One
    sitesym_num, disp = _get_displacement_one(site_symmetry, directions)
    if disp is not None:
        if log_level > 2:
            print("Site symmetry used to expand a direction %s" % disp[0])
            print(site_symmetry[sitesym_num])
        return disp
    # Two
    sitesym_num, disps = _get_displacement_two(site_symmetry, directions)
    if disps is not None:
        if log_level > 2:
            print(
                "Site symmetry used to expand directions %s %s" % (disps[0], disps[1])
            )
            print(site_symmetry[sitesym_num])

        if is_trigonal:
            disps_new = [disps[0]]
            if _is_trigonal_axis(site_symmetry[sitesym_num]):
                if log_level > 2:
                    print("Trigonal axis is found.")
                disps_new.append(np.dot(disps[0], site_symmetry[sitesym_num].T))
                disps_new.append(np.dot(disps_new[1], site_symmetry[sitesym_num].T))
            disps_new.append(disps[1])
            return disps_new
        else:
            return disps
    # Three
    return [directions[0], directions[1], directions[2]]


def _get_displacement_one(site_symmetry, directions=directions_diag):
    """Return one displacement.

    This method tries to find three linearly independent displacements by
    applying site symmetry to an input displacement.

    """
    for direction in directions:
        rot_directions = []
        for r in site_symmetry:
            rot_directions.append(np.dot(direction, r.T))
        num_sitesym = len(site_symmetry)
        for i in range(num_sitesym):
            for j in range(i + 1, num_sitesym):
                det = _determinant(direction, rot_directions[i], rot_directions[j])
                if det != 0:
                    return i, [direction]
    return None, None


def _get_displacement_two(site_symmetry, directions=directions_diag):
    """Return one displacement.

    This method tries to find three linearly independent displacements by
    applying site symmetry to two input displacements.

    """
    for direction in directions:
        rot_directions = []
        for r in site_symmetry:
            rot_directions.append(np.dot(direction, r.T))
        num_sitesym = len(site_symmetry)
        for i in range(num_sitesym):
            for second_direction in directions:
                det = _determinant(direction, rot_directions[i], second_direction)
                if det != 0:
                    return i, [direction, second_direction]
    return None, None


def is_minus_displacement(direction, site_symmetry):
    """Symmetrically check if minus displacement is necessary or not."""
    is_minus = True
    for r in site_symmetry:
        rot_direction = np.dot(direction, r.T)
        if (rot_direction + direction).any():
            continue
        else:
            is_minus = False
            break
    return is_minus


def _is_trigonal_axis(r):
    """Check three folded rotation.

    True if r^3 = identity.

    """
    r3 = np.dot(np.dot(r, r), r)
    if (r3 == np.eye(3, dtype=int)).all():
        return True
    else:
        return False


def _determinant(a, b, c):
    """Return determinant of 3x3 matrix of [a, b, c]."""
    det = (
        a[0] * b[1] * c[2]
        - a[0] * b[2] * c[1]
        + a[1] * b[2] * c[0]
        - a[1] * b[0] * c[2]
        + a[2] * b[0] * c[1]
        - a[2] * b[1] * c[0]
    )
    return det


def get_random_displacements_dataset(
    number_of_snapshots: int,
    num_atoms: int,
    distance: float,
    random_seed: Optional[int] = None,
    is_plusminus: bool = False,
    max_distance: Optional[float] = None,
) -> np.ndarray:
    """Return random displacements at constant displacement distance.

    number_of_snapshots : int,
        Number of snapshots of supercells with random displacements. Random
        displacements are generated displacing all atoms in random directions
        with a fixed displacement distance specified by 'distance' parameter,
        i.e., all atoms in supercell are displaced with the same displacement
        distance in direct space.
    num_atoms : int
        Number of atoms in supercell.
    distance : float
        Displacement distance. Unit is the same as that used for crystal
        structure. In random direction displacements generation with random
        distance, the displacement distance smaller than this value is set to
        this value.
    random_seed : int or None, optional
        Random seed for random displacements generation. Default is None.
    is_plusminus : True, or False, optional
        In addition to sets of usual random displacements for supercell, sets of
        the opposite displacements for supercell are concatenated. Therefore,
        total number of sets of displacements is `2 * number_of_snapshots`.
        Default is False.
    max_distance : float or None, optional
        In random direction and distance displacements generation, this value is
        specified. In random direction and random distance displacements
        generation, this value is used as `max_distance`.

    """
    if np.issubdtype(type(random_seed), np.integer):
        rng = np.random.default_rng(seed=random_seed)
    else:
        rng = np.random.default_rng()

    num_supercells = number_of_snapshots

    if max_distance is None:
        directions = _get_random_directions(num_atoms * num_supercells, rng)
        disps = directions * distance
        supercell_disps = np.array(
            disps.reshape(num_supercells, num_atoms, 3), dtype="double", order="C"
        )
    else:
        if distance > max_distance:
            raise RuntimeError(
                "Random displacements generation failed. max_distance is too small."
            )
        supercell_disps = np.zeros(
            (num_supercells, num_atoms, 3), dtype="double", order="C"
        )
        directions = _get_random_directions(num_atoms * num_supercells, rng).reshape(
            num_supercells, num_atoms, 3
        )
        dists = rng.random(num_supercells) * max_distance
        dists[dists < distance] = distance
        for i, (dirs, dist) in enumerate(zip(directions, dists)):
            supercell_disps[i] = dirs * dist

    if is_plusminus is True:
        supercell_disps = np.array(
            np.concatenate((supercell_disps, -supercell_disps), axis=0),
            dtype="double",
            order="C",
        )
    return supercell_disps


def _get_random_directions(num_atoms: int, rng: np.random.Generator) -> np.ndarray:
    """Return random directions in sphere with radius 1."""
    xy = rng.standard_normal(size=(3, num_atoms * 2))
    r = np.linalg.norm(xy, axis=0)
    condition = r > 1e-10
    return (xy[:, condition][:, :num_atoms] / r[condition][:num_atoms]).T



"""Force constants calculator interfaces."""

# Copyright (C) 2019 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

from collections.abc import Sequence
from typing import Literal, Optional, Union

import numpy as np

from phonopy.exception import ForceCalculatorRequiredError
from phonopy.harmonic.force_constants import FDFCSolver
from phonopy.interface.alm import ALMFCSolver
from phonopy.interface.symfc import SymfcFCSolver
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import Primitive
from phonopy.structure.dataset import get_displacements_and_forces
from phonopy.structure.symmetry import Symmetry

fc_calculator_names = {
    "alm": "ALM",
    "symfc": "symfc",
    "traditional": "phonopy-traditional",
}


def get_fc_solver(
    supercell: PhonopyAtoms,
    dataset: dict,
    primitive: Optional[Primitive] = None,
    fc_calculator: Optional[Literal["traditional", "symfc", "alm"]] = None,
    fc_calculator_options: Optional[str] = None,
    orders: Optional[Sequence[int]] = None,
    is_compact_fc: bool = False,
    symmetry: Optional[Symmetry] = None,
    log_level: int = 0,
) -> FCSolver:
    """Return force constants solver class instance.

    Parameters
    ----------
    supercell : PhonopyAtoms
        Supercell
    dataset : dict
        Dataset that contains displacements, forces, and optionally energies.
    primitive : Primitive
        Primitive cell. Only needed for the traditional FC calculator.
    fc_calculator : str, optional
        Currently 'traditional' (FD method), 'alm', and 'symfc' are supported.
        Default is None, meaning invoking 'traditional'.
    fc_calculator_options : str, optional
        This is arbitrary string.
    orders : Sequence[int], optional
        Orders of force constants to be calculated. Default is None.
    is_compact_fc : bool, optional
        If True, force constants are returned in the compact form.
    symmetry : Symmetry, optional
        Symmetry of supercell. This is used for the traditional and symfc FC
        solver. Default is None.
    log_level : integer or bool, optional
        Verbosity level. False or 0 means quiet. True or 1 means normal level of
        log to stdout. 2 gives verbose mode.

    Returns
    -------
    FCSolver
        Force constants solver class instance.

    """
    fc_solver_name = fc_calculator if fc_calculator is not None else "traditional"
    fc_solver = FCSolver(
        fc_solver_name,
        supercell,
        symmetry=symmetry,
        dataset=dataset,
        is_compact_fc=is_compact_fc,
        primitive=primitive,
        orders=orders,
        options=fc_calculator_options,
        log_level=log_level,
    )
    return fc_solver


def get_fc2(
    supercell: PhonopyAtoms,
    dataset: dict,
    primitive: Optional[Primitive] = None,
    fc_calculator: Optional[Literal["traditional", "symfc", "alm"]] = None,
    fc_calculator_options: Optional[str] = None,
    is_compact_fc: bool = False,
    symmetry: Optional[Symmetry] = None,
    log_level: int = 0,
):
    """Supercell 2nd order force constants (fc2) are calculated.

    The expected shape of supercell fc2 to be returned is
        (len(atom_list), num_atoms, 3, 3).

    Parameters
    ----------
    supercell : PhonopyAtoms
        Supercell
    dataset : dict
        Dataset that contains displacements, forces, and optionally energies.
    primitive : Primitive
        Primitive cell. Only needed for the traditional FC calculator.
    fc_calculator : str, optional
        Currently 'traditional' (FD method), 'alm', and 'symfc' are supported.
        Default is None, meaning invoking 'traditional'.
    fc_calculator_options : str, optional
        This is arbitrary string.
    is_compact_fc : bool, optional
        If True, force constants are returned in the compact form.
    symmetry : Symmetry, optional
        Symmetry of supercell. This is used for the traditional and symfc FC
        solver. Default is None.
    log_level : integer or bool, optional
        Verbosity level. False or 0 means quiet. True or 1 means normal level of
        log to stdout. 2 gives verbose mode.

    Returns
    -------
    fc2 : ndarray
        2nd order force constants. shape=(len(atom_list), num_atoms, 3, 3),
        dtype='double', order='C'.

    """
    fc_solver = get_fc_solver(
        supercell,
        dataset,
        primitive=primitive,
        fc_calculator=fc_calculator,
        fc_calculator_options=fc_calculator_options,
        orders=[2],
        is_compact_fc=is_compact_fc,
        symmetry=symmetry,
        log_level=log_level,
    )
    return fc_solver.force_constants[2]


class FCSolver:
    """Force constants calculator interface."""

    def __init__(
        self,
        fc_solver_name: Literal["traditional", "symfc", "alm"],
        supercell: PhonopyAtoms,
        symmetry: Optional[Symmetry] = None,
        dataset: Optional[dict] = None,
        is_compact_fc: bool = False,
        primitive: Optional[Primitive] = None,
        orders: Optional[Sequence[int]] = None,
        options: Optional[str] = None,
        log_level: int = 0,
    ):
        """Init method.

        Force constants are calculated if necessary data are provided.

        For the symfc FC solver, SymfcFCSolver instance can be returned without
        necessary data for computing force constants. For the other solvers,
        raise an error.

        Parameters
        ----------
        fc_solver_name : Literal["traditional", "symfc", "alm"]
            Force constants calculator name.
        supercell : PhonopyAtoms
            Supercell.
        symmetry : Symmetry, optional
            Symmetry of supercell. Default is None.
        dataset : dict, optional
            Dataset that contains displacements, forces, and optionally
            energies. Default is None.
        is_compact_fc : bool, optional
            If True, force constants are returned in the compact form.
        primitive : Primitive, optional
            Primitive cell. This is needed for the traditional and alm FC
            calculator. Default is None.
        orders : Sequence[int], optional
            Orders of force constants to be calculated. Default is None.
        options : str, optional
            This is arbitrary string that is used for each FC solver. Default is
            None.
        log_level : int, optional
            Log level. Default is 0.

        """
        self._supercell = supercell
        self._symmetry = symmetry
        self._dataset = dataset
        self._is_compact_fc = is_compact_fc
        self._primitive = primitive
        self._orders = orders
        self._options = options
        self._log_level = log_level

        self._fc_solver = self._set_fc_solver(fc_solver_name)

    @property
    def fc_solver(self) -> Union[FDFCSolver, SymfcFCSolver, ALMFCSolver]:
        """Return force constants solver class instance."""
        return self._fc_solver

    @property
    def force_constants(self) -> dict[int, np.ndarray]:
        """Return force constants."""
        return self._fc_solver.force_constants

    def _set_fc_solver(self, fc_calculator_name: str):
        if fc_calculator_name not in fc_calculator_names:
            raise ValueError(
                f"fc_calculator_name must be one of {fc_calculator_names.keys()}"
            )

        if fc_calculator_name == "traditional":
            return self._set_traditional_solver()
        if fc_calculator_name == "symfc":
            return self._set_symfc_solver()
        if fc_calculator_name == "alm":
            return self._set_alm_solver()

    def _set_traditional_solver(self, solver_class: Optional[type] = FDFCSolver):
        if self._primitive is None:
            raise RuntimeError(
                "Primitive cell is required for the traditional FC solver."
            )

        if self._dataset is None:
            raise RuntimeError(
                "Displacement-force dataset is required for the traditional FC solver."
            )

        if "displacements" in self._dataset:
            lines = [
                "Type-II dataset for displacements and forces was provided, ",
                "but the selected force constants calculator cannot process it.",
                "Use another force constants calculator, e.g., symfc, ",
                "to generate force constants.",
            ]
            raise ForceCalculatorRequiredError("\n".join(lines))

        return solver_class(
            self._supercell,
            self._primitive,
            self._symmetry,
            self._dataset,
            is_compact_fc=self._is_compact_fc,
            log_level=self._log_level,
        )

    def _set_symfc_solver(self):
        from phonopy.interface.symfc import SymfcFCSolver, parse_symfc_options

        if self._dataset is None:
            return SymfcFCSolver(
                self._supercell,
                symmetry=self._symmetry,
                options=parse_symfc_options(self._options),
                is_compact_fc=self._is_compact_fc,
                log_level=self._log_level,
            )
        else:
            displacements, forces = self._get_displacements_and_forces()
            symfc_solver = SymfcFCSolver(
                self._supercell,
                displacements,
                forces,
                symmetry=self._symmetry,
                options=parse_symfc_options(self._options),
                is_compact_fc=self._is_compact_fc,
                log_level=self._log_level,
            )
            if self._orders is not None:
                symfc_solver.run(orders=self._orders)
            return symfc_solver

    def _set_alm_solver(self):
        from phonopy.interface.alm import ALMFCSolver

        if self._primitive is None:
            raise RuntimeError(
                "Primitive cell is required for the traditional FC solver."
            )

        if self._dataset is None:
            raise RuntimeError(
                "Displacement-force dataset is required for ALM FC solver."
            )
        displacements, forces = self._get_displacements_and_forces()

        return ALMFCSolver(
            self._supercell,
            self._primitive,
            displacements,
            forces,
            max(self._orders) - 1,
            is_compact_fc=self._is_compact_fc,
            options=self._options,
            log_level=self._log_level,
        )

    def _get_displacements_and_forces(self):
        """Return displacements and forces for fc2."""
        return get_displacements_and_forces(self._dataset)



# Copyright (C) 2018 Antti Karttunen (antti.j.karttunen@iki.fi)
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys


def get_options():
    """Parse options."""
    import argparse

    parser = argparse.ArgumentParser(
        description="Phonopy crystal-born command-line-tool"
    )
    parser.add_argument(
        "filename", nargs="*", help="Filename: CRYSTAL output file (default: crystal.o)"
    )
    args = parser.parse_args()
    return args


def read_crystal_epsilon(epscol, epslines):
    """Read epsilon in CRYSTAL14 format."""
    # XX, XY, XZ, YY, YZ, ZZ are given (6 rows). YX = XY, ZX = XZ, ZY = YZ
    # Format for CRYSTAL17 (epscol == 4):
    # 0.000   XX      57.1760       0.0000       3.2117       0.0000       2.2117
    # Format for CRYSTAL14 (epscol == 3):
    # XX      57.1760        0.0000        3.2117        2.2117
    eps = []  # Will have 9 elements 0..8
    eps.append(float(epslines[0].split()[epscol]))  # XX (0)
    eps.append(float(epslines[1].split()[epscol]))  # XY (1)
    eps.append(float(epslines[2].split()[epscol]))  # XZ (2)
    eps.append(eps[1])  # YX = XY (3)
    eps.append(float(epslines[3].split()[epscol]))  # YY (4)
    eps.append(float(epslines[4].split()[epscol]))  # YZ (5)
    eps.append(eps[2])  # ZX = XZ (6)
    eps.append(eps[5])  # ZY = YZ (7)
    eps.append(float(epslines[5].split()[epscol]))  # ZZ (8)
    return eps


def run():
    """Run phonopy-crystal-born."""
    args = get_options()

    if args.filename:
        crystal_filename = args.filename[0]
    else:
        crystal_filename = "crystal.o"

    # Read in the output file
    try:
        with open(crystal_filename, "r") as crystal_file:
            lines = crystal_file.readlines()
    except OSError:
        print(
            "CRYSTAL output file {} cannot be opened for reading".format(
                crystal_filename
            )
        )
        sys.exit(1)

    # Recommended CRYSTAL calculation type:
    # Gamma-point FREQCALC with INTCPHF and INTENS
    # The file will include both dielectric tensor and effective Born charges
    epsilon = []
    zeff = []
    ll = 0
    while ll < len(lines):
        line = lines[ll]
        # Parse the information about atoms
        if "PRIMITIVE CELL - CENTRING CODE" in line:
            ll += 4
            # ATOMS IN THE ASYMMETRIC UNIT    2 - ATOMS IN THE UNIT CELL:    6
            N_asym_atoms = int(lines[ll].split()[5])
            N_atoms = int(lines[ll].split()[12])
            ll += 3
            # Check for each atom if it belongs to the asymmetric unit
            # 1 T  22 TI    4.721218104494E-21  3.307446203077E-21  1.413771901417E-21
            is_asym = []
            for _ in range(0, N_atoms):
                atomdata = lines[ll].split()
                is_asym.append(atomdata[1])
                ll += 1

        # Parse dielectric tensor from INTCPHF
        elif "FR.(eV) COMP.   ALPHA(Re,Im)           EPSILON(Re,Im)" in line:
            # CRYSTAL17
            epsilon = read_crystal_epsilon(4, lines[ll + 1 : ll + 7])
            ll += 7
        elif "COMPONENT    ALPHA(REAL, IMAGINARY)         EPSILON       CHI(1)" in line:
            epsilon = read_crystal_epsilon(3, lines[ll + 1 : ll + 7])
            ll += 7

        # Parse Born charges
        elif "ATOMIC BORN CHARGE TENSOR" in line:
            ll += 6
            for atom in range(0, N_atoms):
                if is_asym[atom] == "T":
                    zeffatom = []
                    for _ in range(0, 3):
                        # 1     8.3860E-01  3.4861E-01  3.4861E-01
                        zeffatom += [float(x) for x in lines[ll].split()[1:4]]
                        ll += 1
                    zeff.append(zeffatom)
                    ll += 4
                else:
                    ll += 7
        ll += 1  # while l < len(lines)

    # Output BORN if epsilon and zeff are available
    if len(epsilon) == 9 and len(zeff) == N_asym_atoms:
        # Conversion factor
        bornlines = "default\n"
        # Dielectric tensor
        bornlines += ("%6.4f " * 9 + "\n") % tuple(epsilon)
        # Effective charges
        for atom in range(0, N_asym_atoms):
            bornlines += ("%6.4f " * 9 + "\n") % tuple(zeff[atom])
        print(bornlines, end="")



"""API of mode Grueneisen parameter calculation."""

# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from phonopy.gruneisen.band_structure import GruneisenBandStructure
from phonopy.gruneisen.mesh import GruneisenMesh


class PhonopyGruneisen:
    """Class to calculate mode Grueneisen parameters."""

    def __init__(self, phonon, phonon_plus, phonon_minus, delta_strain=None):
        """Init method.

        Parameters
        ----------
        phonon, phonon_plus, phonon_minus : Phonopy
            Phonopy instances of the same crystal with differet volumes,
            V_0, V_0 + dV, V_0 - dV.
        delta_strain : float, optional
            Default is None, which gives dV / V_0.

        """
        self._phonon = phonon
        self._phonon_plus = phonon_plus
        self._phonon_minus = phonon_minus
        self._delta_strain = delta_strain

        self._mesh = None
        self._band_structure = None

    def get_phonon(self):
        """Return Phonopy class instance at dV=0."""
        return self._phonon

    def set_mesh(
        self,
        mesh,
        shift=None,
        is_time_reversal=True,
        is_gamma_center=False,
        is_mesh_symmetry=True,
    ):
        """Set sampling mesh."""
        for phonon in (self._phonon, self._phonon_plus, self._phonon_minus):
            if phonon.dynamical_matrix is None:
                print("Warning: Dynamical matrix has not yet built.")
                return False

        symmetry = phonon.primitive_symmetry
        rotations = symmetry.pointgroup_operations
        self._mesh = GruneisenMesh(
            self._phonon.dynamical_matrix,
            self._phonon_plus.dynamical_matrix,
            self._phonon_minus.dynamical_matrix,
            mesh,
            delta_strain=self._delta_strain,
            shift=shift,
            is_time_reversal=is_time_reversal,
            is_gamma_center=is_gamma_center,
            is_mesh_symmetry=is_mesh_symmetry,
            rotations=rotations,
            factor=self._phonon.unit_conversion_factor,
        )
        return True

    def get_mesh(self):
        """Return mode Grueneisen parameters calculated on sampling mesh."""
        if self._mesh is None:
            return None
        else:
            return (
                self._mesh.get_qpoints(),
                self._mesh.get_weights(),
                self._mesh.get_frequencies(),
                self._mesh.get_eigenvectors(),
                self._mesh.get_gruneisen(),
            )

    def write_yaml_mesh(self, filename="gruneisen_mesh.yaml"):
        """Write mesh sampling calculation results to file in yaml."""
        self._mesh.write_yaml(filename=filename)

    def write_hdf5_mesh(self, filename="gruneisen_mesh.hdf5"):
        """Write mesh sampling calculation results to file in hdf5."""
        self._mesh.write_hdf5(filename=filename)

    def plot_mesh(
        self, cutoff_frequency=None, color_scheme=None, marker="o", markersize=None
    ):
        """Return pyplot of mesh sampling calculation results."""
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        ax.xaxis.set_ticks_position("both")
        ax.yaxis.set_ticks_position("both")
        ax.xaxis.set_tick_params(which="both", direction="in")
        ax.yaxis.set_tick_params(which="both", direction="in")
        self._mesh.plot(
            plt,
            cutoff_frequency=cutoff_frequency,
            color_scheme=color_scheme,
            marker=marker,
            markersize=markersize,
        )
        return plt

    def set_band_structure(self, bands):
        """Set band structure paths."""
        self._band_structure = GruneisenBandStructure(
            bands,
            self._phonon.dynamical_matrix,
            self._phonon_plus.dynamical_matrix,
            self._phonon_minus.dynamical_matrix,
            delta_strain=self._delta_strain,
            factor=self._phonon.unit_conversion_factor,
        )

    def get_band_structure(self):
        """Return band structure calculation results."""
        band = self._band_structure
        return (
            band.get_qpoints(),
            band.get_distances(),
            band.get_frequencies(),
            band.get_eigenvectors(),
            band.get_gruneisen(),
        )

    def write_yaml_band_structure(self, filename="gruneisen_band.yaml"):
        """Write band structure calculation results to file in yaml."""
        self._band_structure.write_yaml(filename=filename)

    def plot_band_structure(self, epsilon=1e-4, color_scheme=None):
        """Return pyplot of band structure calculation results."""
        import matplotlib.pyplot as plt

        fig, axarr = plt.subplots(2, 1)
        for ax in axarr:
            ax.xaxis.set_ticks_position("both")
            ax.yaxis.set_ticks_position("both")
            ax.xaxis.set_tick_params(which="both", direction="in")
            ax.yaxis.set_tick_params(which="both", direction="in")
            self._band_structure.plot(axarr, epsilon=epsilon, color_scheme=color_scheme)
        return plt



# Copyright (C) 2018 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from phonopy.interface.vasp import parse_vasprunxml
from phonopy.qha.electron import get_free_energy_at_T

"""Calculate electronic free energy from vasprun.xml at temperatures

Here the free energy is approximately given by:

    energy(sigma->0) - energy(T=0) + energy(T) - entropy(T) * T

"""


def get_options():
    """Parse command-line options."""
    import argparse

    parser = argparse.ArgumentParser(description="Phonopy vasp-efe command-line-tool")
    parser.set_defaults(tmax=1000.0, tmin=0.0, tstep=10.0)
    parser.add_argument(
        "--tmax", dest="tmax", type=float, help="Maximum calculated temperature"
    )
    parser.add_argument(
        "--tmin", dest="tmin", type=float, help="Minimum calculated temperature"
    )
    parser.add_argument(
        "--tstep", dest="tstep", type=float, help="Calculated temperature step"
    )
    parser.add_argument(
        "filenames",
        nargs="*",
        help="Filenames: vasprun.xml's of all volumes in correct order",
    )
    args = parser.parse_args()
    return args


def get_free_energy_lines(temperatures, free_energies):
    """Return Free energy lines."""
    lines = []
    n_vol = free_energies.shape[1]
    for t, fe in zip(temperatures, free_energies):
        lines.append(("%10.4f " + " %15.8f" * n_vol) % ((t,) + tuple(fe)))
    return lines


def get_fe_ev_lines(args):
    """Return Free energy vs volume lines."""
    volumes = []
    energy_sigma0 = []
    free_energies = []
    temperatures = None
    for filename in args.filenames:
        vxml = parse_vasprunxml(filename)
        weights = vxml.k_weights
        eigenvalues = vxml.eigenvalues[:, :, :, 0]
        n_electrons = vxml.NELECT
        energy = vxml.energies[-1, 1]
        temps, fe = get_free_energy_at_T(
            args.tmin, args.tmax, args.tstep, eigenvalues, weights, n_electrons
        )
        volumes.append(vxml.volume[-1])
        energy_sigma0.append(energy)
        free_energies.append(energy - fe[0] + fe)
        if temperatures is None:
            temperatures = temps
        else:
            assert (np.abs(temperatures - temps) < 1e-5).all()

    lines_fe = []
    lines_fe.append(("# volume:  " + " %15.8f" * len(volumes)) % tuple(volumes))
    lines_fe.append("#    T(K)     Free energies")
    lines_fe += get_free_energy_lines(temperatures, np.transpose(free_energies))

    lines_ev = ["#   cell volume        energy of cell other than phonon"]
    lines_ev += ["%20.8f %20.8f" % (v, e) for v, e in zip(volumes, energy_sigma0)]

    return lines_fe, lines_ev


def run():
    """Run phonopy-vasp-efe."""
    args = get_options()

    lines_fe, lines_ev = get_fe_ev_lines(args)

    with open("fe-v.dat", "w") as w:
        w.write("\n".join(lines_fe))
        w.write("\n")
        print('* Electronic free energies are written in "fe-v.dat".')

    with open("e-v.dat", "w") as w:
        w.write("\n".join(lines_ev))
        w.write("\n")
        print('* energy (sigma->0) and volumes are written in "e-v.dat".')



"""Example of calculation of irreps of MgB2."""

import numpy as np

import phonopy

phonon = phonopy.load(unitcell_filename="POSCAR-unitcell", supercell_matrix=[3, 3, 2])
print("Space group: %s" % phonon.symmetry.get_international_table())

# Character table
phonon.set_irreps([1.0 / 3, 1.0 / 3, 0], 1e-4)
ct = phonon.get_irreps()
band_indices = ct.get_band_indices()
characters = np.rint(ct.get_characters()).real
for bi, cts in zip(band_indices, characters):
    print("%s %s" % (np.array(bi) + 1, cts))
# phonon.show_character_table()



"""Example to obtain PhonopyYaml instance."""

import phonopy
from phonopy.interface.phonopy_yaml import PhonopyYaml

phonon = phonopy.load(
    supercell_matrix=[[2, 0, 0], [0, 2, 0], [0, 0, 2]],
    primitive_matrix=[[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]],
    unitcell_filename="POSCAR-unitcell",
    force_sets_filename="FORCE_SETS",
    born_filename="BORN",
)
phpy_yaml = PhonopyYaml(calculator="vasp", settings={"force_constants": True})
phpy_yaml.set_phonon_info(phonon)
print(phpy_yaml)



"""Mode Grueneisen parameter calculation."""

# Copyright (C) 2012 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import Union

import numpy as np

from phonopy.harmonic.dynamical_matrix import DynamicalMatrix, DynamicalMatrixNAC
from phonopy.phonon.band_structure import estimate_band_connection
from phonopy.phonon.degeneracy import rotate_eigenvectors


class GruneisenBase:
    """Base class of mode Grueneisen parameter calculation classes."""

    def __init__(
        self,
        dynmat: Union[DynamicalMatrix, DynamicalMatrixNAC],
        dynmat_plus: Union[DynamicalMatrix, DynamicalMatrixNAC],
        dynmat_minus: Union[DynamicalMatrix, DynamicalMatrixNAC],
        delta_strain=None,
        qpoints=None,
        is_band_connection=False,
    ):
        """Init method."""
        self._dynmat = dynmat
        self._dynmat_plus = dynmat_plus
        self._dynmat_minus = dynmat_minus
        if delta_strain is None:
            volume = dynmat.primitive.volume
            volume_plus = dynmat_plus.primitive.volume
            volume_minus = dynmat_minus.primitive.volume
            dV = volume_plus - volume_minus
            self._delta_strain = dV / volume
        else:
            self._delta_strain = delta_strain
        self._is_band_connection = is_band_connection
        self._qpoints = qpoints

        self._gruneisen = None
        self._eigenvalues = None
        if qpoints is not None:
            self._set_gruneisen()

    def set_qpoints(self, qpoints):
        """Set q-points."""
        self._qpoints = qpoints
        self._set_gruneisen()

    def get_gruneisen(self):
        """Return mode Grueneisen parameters."""
        return self._gruneisen

    def get_eigenvalues(self):
        """Return eigenvalues."""
        return self._eigenvalues

    def get_eigenvectors(self):
        """Return eigenvectors."""
        return self._eigenvectors

    def _set_gruneisen(self):
        if self._is_band_connection:
            self._q_direction = self._qpoints[0] - self._qpoints[-1]
            band_order = range(len(self._dynmat.primitive) * 3)
            prev_eigvecs = None

        edDe = []  # <e|dD|e>
        eigvals = []
        eigvecs = []
        for _, q in enumerate(self._qpoints):
            if self._is_band_connection and isinstance(
                self._dynmat, DynamicalMatrixNAC
            ):
                self._dynmat.run(q, q_direction=self._q_direction)
            else:
                self._dynmat.run(q)

            dm = self._dynmat.dynamical_matrix
            evals, evecs = np.linalg.eigh(dm)
            evals_at_q = evals.real
            dD = self._get_dD(q, self._dynmat_minus, self._dynmat_plus)
            evecs_at_q, edDe_at_q = rotate_eigenvectors(evals_at_q, evecs, dD)

            if self._is_band_connection:
                if prev_eigvecs is not None:
                    band_order = estimate_band_connection(
                        prev_eigvecs, evecs_at_q, band_order
                    )
                eigvals.append(evals_at_q[band_order])
                eigvecs.append(evecs_at_q[:, band_order])
                edDe.append(edDe_at_q[band_order])
                prev_eigvecs = evecs_at_q
            else:
                eigvals.append(evals_at_q)
                eigvecs.append(evecs_at_q)
                edDe.append(edDe_at_q)

        edDe = np.array(edDe, dtype="double", order="C")
        self._eigenvalues = np.array(eigvals, dtype="double", order="C")
        itemsize = self._eigenvalues.itemsize
        self._eigenvectors = np.array(
            eigvecs, dtype=("c%d" % (itemsize * 2)), order="C"
        )
        self._gruneisen = -edDe / self._delta_strain / self._eigenvalues / 2

    def _get_dD(
        self,
        q,
        d_a: Union[DynamicalMatrix, DynamicalMatrixNAC],
        d_b: Union[DynamicalMatrix, DynamicalMatrixNAC],
    ):
        if (
            self._is_band_connection
            and isinstance(d_a, DynamicalMatrixNAC)
            and isinstance(d_b, DynamicalMatrixNAC)
        ):
            d_a.run(q, q_direction=self._q_direction)
            d_b.run(q, q_direction=self._q_direction)
        else:
            d_a.run(q)
            d_b.run(q)
        dm_a = d_a.dynamical_matrix
        dm_b = d_b.dynamical_matrix
        return dm_b - dm_a



"""Band unfolding calculation."""

# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings

import numpy as np

from phonopy import Phonopy
from phonopy.harmonic.dynmat_to_fc import get_commensurate_points
from phonopy.harmonic.force_constants import compact_fc_to_full_fc
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import get_supercell


class Unfolding:
    """Calculation of a phonon unfolding method.

    Implementation of an unfolding method by P. B. Allen et al., Phys. Rev. B
    87, 085322 (2013)

    T(r_i) in this implementation is defined as

        T(r_i) f(x) = f(x - r_i).

    The sign is opposite from that written in the Allen's paper. Bloch wave is
    defined in the same way for phase convention

        Psi_k(x + r) = exp(ikr) Psi_k(x).

    By these, sign of phase in Eq.(3) (Eq.(7) as well) is opposite.

    Vacancies are treated as atomic sites where no atoms exist. Interstitials
    are treated as atomic sites with the interstitial atoms in a specific
    virtual primitive cell, but the respective atomic sites in the other virtual
    primitive cells are set as vacancies.

    The unfolded band structure may be plotted as
        x (wave vector): qpoints y (frequency): frequencies z (intencity) :
        unfolding_weights


    Attributes
    ----------
    unfolding_weights : ndarray
        Unfolding weights. shape=(qpoints in supercell, supercell atoms * 3),
        dtype='double', order='C'
    frequencies : ndarray
        Phonon frequencies of the supercell. By unfolding, these are considered
        as the phonon frequencies at the specified qpoints but with the
        unfolding weights. shape=(qpoints in supercell, supercell atoms * 3),
        dtype='double', order='C'
    commensurate_points : ndarray
        Commensurate points corresponding to ``supercell_matrix``. shape=(N, 3),
        dtype='double', order='C' where N = det(supercell_matrix)
    qpoints : ndarray
        q-points in supercell BZ.

    """

    def __init__(
        self, phonon: Phonopy, supercell_matrix, ideal_positions, atom_mapping, qpoints
    ):
        """Init method.

        Parameters
        ----------
        phonon : Phonopy
            Phonopy object to be unfolded.
        supercell_matrix : array_like
            Matrix that represents the virtual primitive translation enforced
            inside the supercell. This works like an inverse primitive matrix.
            shape=(3, 3), dtype='intc'
        ideal_positions : array_like
            Positions of atomic sites in supercell. This corresponds to those
            in a set of virtual primitive cells in the supercell.
            shape=(3, 3), dtype='double'
        atom_mapping : list
            Atomic index mapping from ``ideal_positions`` to supercell atoms
            in ``phonon``. The elements of this list are intergers for atoms
            and None for vacancies.
        qpoints : array_like
            q-points in reciprocal virtual-primitive-cell coordinates
            shape=(num_qpoints, 3), dtype='double'

        """
        self._phonon = self._get_supercell_phonon(phonon)
        self._supercell_matrix = np.array(supercell_matrix, dtype="intc")
        self._ideal_positions = np.array(ideal_positions, dtype="double")
        self._qpoints_p = qpoints  # in PBZ
        self._qpoints_s = self._get_qpoints_in_SBZ()  # in SBZ
        self._symprec = self._phonon.symmetry.tolerance

        self._frequencies = None
        self._eigvecs = None  # This may have anormal array shape.
        self._unfolding_weights = None
        self._q_count = None  # As counter for iterator

        # Commensurate q-vectors in PBZ
        self._comm_points = get_commensurate_points(self._supercell_matrix)

        self._trans_s = None  # in SC (see docstring in _set_translations)
        self._trans_p = None  # in PC (see docstring in _set_translations)
        self._N = None
        self._set_translations()

        self._atom_mapping = None
        self._index_map_inv = None
        self._set_index_map(atom_mapping)

    def __iter__(self):
        """Define iterator over q-points."""
        return self

    def run(self):
        """Calculate band unfolding at all q-points."""
        self.prepare()
        for _ in self:
            pass

    def __next__(self):
        """Calculate band unfolding at a q-point."""
        if self._q_count == len(self._qpoints_s):
            self._eigvecs = None
            raise StopIteration

        self._solve_phonon()
        self._unfolding_weights[self._q_count] = self._get_unfolding_weights()
        self._q_count += 1
        return self._unfolding_weights[self._q_count - 1]

    def prepare(self):
        """Set up calculation."""
        self._q_count = 0
        self._unfolding_weights = np.zeros(
            (len(self._qpoints_s), len(self._phonon.supercell) * 3), dtype="double"
        )
        self._frequencies = np.zeros_like(self._unfolding_weights)

    @property
    def commensurate_points(self):
        """Return commensurate points.

        See details in Attributes.

        """
        return self._comm_points

    def get_commensurate_points(self):
        """Return commensurate points."""
        warnings.warn(
            "Unfolding.get_commensurate_points() is deprecated. "
            "Use commensurate_points attribute instaed.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.commensurate_points

    @property
    def unfolding_weights(self):
        """Return unfolding weights.

        See details in Attributes.

        """
        return self._unfolding_weights

    def get_unfolding_weights(self):
        """Return unfolding weights."""
        warnings.warn(
            "Unfolding.get_unfolding_weights() is deprecated. "
            "Use unfolding_weights attribute instaed.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.unfolding_weights

    @property
    def frequencies(self):
        """Return phonon frequencies.

        See details in Attributes.

        """
        return self._frequencies

    def get_frequencies(self):
        """Return phonon frequencies."""
        warnings.warn(
            "Unfolding.get_frequencies() is deprecated. "
            "Use frequencies attribute instaed.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.frequencies

    @property
    def qpoints(self):
        """Return q-points in supercell BZ.

        See details in Attributes.

        """
        return self._qpoints_s

    def _get_qpoints_in_SBZ(self):
        qpoints = np.dot(self._qpoints_p, self._supercell_matrix)
        qpoints -= np.rint(qpoints)
        return qpoints

    def _set_translations(self):
        """Set primitive translations in supercell.

        _trans_s
            Translations with respect to supercell basis vectors
        _trans_p
            Translations with respect to primitive cell basis vectors
        _N
            Number of the translations = det(supercel_matrix)

        """
        pcell = PhonopyAtoms(
            symbols=["H"], scaled_positions=[[0, 0, 0]], cell=np.diag([1, 1, 1])
        )
        smat = self._supercell_matrix
        self._trans_s = get_supercell(pcell, smat).scaled_positions
        self._trans_p = np.dot(self._trans_s, self._supercell_matrix.T)
        self._N = len(self._trans_s)

    def _set_index_map(self, atom_mapping):
        """Set atomic site index mapping.

        T(r_i)^-1 in Eq.(3) is given as permutation of atom indices.

        _index_map_inv : ndarray
            For each translation (shift), atomic indices of the positions by
            inverse translation (_ideal_positions - shift) are searched and
            stored, i.e.,

                (shift, atom): pos[atom] - shift -> atom'

            The indices are used to select eigenvectors, by which T(r_i)|KJ> is
            represented.
            shape=(num_trans, num_sites), dtype='intc'

        """
        lattice = self._phonon.supercell.cell
        natom = len(self._ideal_positions)
        index_map_inv = np.zeros((self._N, natom), dtype="intc")
        for i, shift in enumerate(self._trans_s):
            for j, p in enumerate(self._ideal_positions - shift):  # minus r_i
                diff = self._ideal_positions - p
                diff -= np.rint(diff)
                dist = np.sqrt((np.dot(diff, lattice) ** 2).sum(axis=1))
                # k is index in _ideal_positions.
                k = np.where(dist < self._symprec)[0][0]
                index_map_inv[i, j] = k
        self._index_map_inv = index_map_inv

        self._atom_mapping = np.zeros(len(atom_mapping), dtype="int")
        for i, idx in enumerate(atom_mapping):
            if idx is None:
                self._atom_mapping[i] = -1
            else:
                self._atom_mapping[i] = idx

    def _solve_phonon(self):
        self._phonon.run_qpoints(self._qpoints_s[self._q_count], with_eigenvectors=True)
        qpt = self._phonon.get_qpoints_dict()
        self._frequencies[self._q_count] = qpt["frequencies"][0]
        eigvecs = qpt["eigenvectors"][0]

        if -1 in self._atom_mapping:
            shape = list(eigvecs.shape)
            shape[0] += 3
            self._eigvecs = np.zeros(tuple(shape), dtype=eigvecs.dtype)
            self._eigvecs[:-3, :] = eigvecs
        else:
            self._eigvecs = eigvecs

    def _get_unfolding_weights(self):
        """Calculate Eq. (7).

        k = K + G + g

        K -> _qpoints_s[_q_index] (in SBZ)
        k -> _qpoints_p[_q_index] (in PBZ)
        G -> _comm_points (in PBZ)
        g -> a reciprocal lattice point in PBZ (unused explicitly)
        j -> Primitive translations in supercell (_trans_p)
        J -> Band indices of supercell phonon modes (axis=1 or eigvecs)

        The phase factor corresponding to K is not included in eigvecs
        with our choice of dynamical matrix.

        """
        eigvecs = self._eigvecs
        dtype = "c%d" % (np.dtype("double").itemsize * 2)
        q_p = self._qpoints_p[self._q_count]  # k
        q_s = self._qpoints_s[self._q_count]  # K
        diff = q_p - np.dot(q_s, np.linalg.inv(self._supercell_matrix))

        # Search G points corresponding to k = G + K
        for G in self._comm_points:
            d = diff - G
            d -= np.rint(d)
            if (np.abs(d) < 1e-5).all():
                break

        e = np.zeros((len(self._atom_mapping) * 3, eigvecs.shape[1]), dtype=dtype)
        phases = np.exp(2j * np.pi * np.dot(self._trans_p, G))

        for phase, indices in zip(phases, self._atom_mapping[self._index_map_inv]):
            eig_indices = (np.c_[indices * 3, indices * 3 + 1, indices * 3 + 2]).ravel()
            e += eigvecs[eig_indices, :] * phase
        e /= self._N
        weights = (e.conj() * e).sum(axis=0)

        # indices = self._atom_mapping
        # eig_indices_r = (
        #     np.c_[indices * 3, indices * 3 + 1, indices * 3 + 2]).ravel()
        # for phase, indices in zip(
        #         phases, self._atom_mapping[self._index_map_inv]):
        #     eig_indices_l = (
        #         np.c_[indices * 3, indices * 3 + 1, indices * 3 + 2]).ravel()
        #     e += eigvecs[eig_indices_l, :] * eigvecs[eig_indices_r, :].conj() * phase
        # e /= self._N
        # weights = e.sum(axis=0)

        if (weights.imag > 1e-5).any():
            print("Phonopy warning: Encountered imaginary values.")

        return weights.real

    def _get_supercell_phonon(self, ph_in: Phonopy):
        """Return Phonopy instance of supercell as the primitive."""
        ph = Phonopy(ph_in.supercell, supercell_matrix=[1, 1, 1], primitive_matrix="P")
        fc_shape = ph_in.force_constants.shape
        if fc_shape[0] == fc_shape[1]:  # assume full fc
            ph.force_constants = ph_in.force_constants.copy()
        else:
            ph.force_constants = compact_fc_to_full_fc(ph_in, ph_in.force_constants)

        if ph_in.nac_params:
            p2p = ph_in.primitive.p2p_map
            s2p = ph_in.primitive.s2p_map
            s2pp = [p2p[i] for i in s2p]
            born_in = ph_in.nac_params["born"]
            born = [born_in[i] for i in s2pp]
            nac_params = {
                "born": np.array(born, dtype="double", order="C"),
                "factor": ph_in.nac_params["factor"],
                "dielectric": ph_in.nac_params["dielectric"].copy(),
            }
            ph.nac_params = nac_params
        return ph



"""Routines to handle various calculator interfaces."""

# Copyright (C) 2014 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import pathlib
import warnings
from argparse import ArgumentParser
from collections.abc import Sequence
from math import pi, sqrt
from typing import Optional, Union

import numpy as np
import yaml

from phonopy.file_IO import get_supported_file_extensions_for_compression
from phonopy.interface.phonopy_yaml import PhonopyYaml
from phonopy.interface.vasp import sort_positions_by_symbols
from phonopy.physical_units import get_physical_units
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import determinant
from phonopy.structure.dataset import get_displacements_and_forces

calculator_info = {
    "abacus": {"option": {"name": "--abacus", "help": "Invoke ABACUS mode"}},
    "abinit": {"option": {"name": "--abinit", "help": "Invoke Abinit mode"}},
    "aims": {"option": {"name": "--aims", "help": "Invoke FHI-aims mode"}},
    "castep": {"option": {"name": "--castep", "help": "Invoke CASTEP mode"}},
    "cp2k": {"option": {"name": "--cp2k", "help": "Invoke CP2K mode"}},
    "crystal": {"option": {"name": "--crystal", "help": "Invoke CRYSTAL mode"}},
    "dftbp": {"option": {"name": "--dftb+", "help": "Invoke dftb+ mode"}},
    "elk": {"option": {"name": "--elk", "help": "Invoke elk mode"}},
    "fleur": {"option": {"name": "--fleur", "help": "Invoke Fleur mode"}},
    "lammps": {"option": {"name": "--lammps", "help": "Invoke Lammps mode"}},
    "qe": {"option": {"name": "--qe", "help": "Invoke Quantum espresso (QE) mode"}},
    "siesta": {"option": {"name": "--siesta", "help": "Invoke Siesta mode"}},
    "turbomole": {"option": {"name": "--turbomole", "help": "Invoke TURBOMOLE mode"}},
    "vasp": {"option": {"name": "--vasp", "help": "Invoke Vasp mode"}},
    "wien2k": {"option": {"name": "--wien2k", "help": "Invoke Wien2k mode"}},
    "pwmat": {"option": {"name": "--pwmat", "help": "Invoke PWmat mode"}},
}


def add_arguments_of_calculators(parser: ArgumentParser, calculator_info):
    """Add options of calculators to ArgumentParser class instance."""
    for calculator in calculator_info:
        option = calculator_info[calculator]["option"]
        parser.add_argument(
            option["name"],
            dest="%s_mode" % calculator,
            action="store_true",
            default=False,
            help=option["help"],
        )


def get_interface_mode(args_dict):
    """Return calculator name.

    The calculator name is obtained from command option arguments where
    argparse is used. The argument attribute name has to be
    "{calculator}_mode". Then this method returns "{calculator}".

    """
    for calculator in calculator_info:
        mode = "%s_mode" % calculator
        if mode in args_dict and args_dict[mode]:
            return calculator
    return None


def convert_crystal_structure(filename_in, interface_in, filename_out, interface_out):
    """Convert crystal structures between different calculator interfaces."""
    cell, _ = read_crystal_structure(filename=filename_in, interface_mode=interface_in)
    units_in = get_calculator_physical_units(interface_in)
    units_out = get_calculator_physical_units(interface_out)
    factor = units_in["distance_to_A"] / units_out["distance_to_A"]
    cell.cell = cell.cell * factor
    write_crystal_structure(filename_out, cell, interface_mode=interface_out)


def write_crystal_structure(
    filename: Optional[str],
    cell: PhonopyAtoms,
    interface_mode: Optional[str] = None,
    optional_structure_info: Optional[tuple] = None,
):
    """Write crystal structure to file in each calculator format.

    filename : str, optional
        File name to be used to write out the crystal structure.
    cell : PhonopyAtoms
        Crystal structure
    interface_mode : str, optional
        Calculator interface such as 'vasp', 'qe', ... Default is None,
        that is equivalent to 'vasp'.
    optional_structure_info : tuple, optional
        Information returned by the method ``read_crystal_structure``.
        See the docstring. Default is None.

    """
    if interface_mode is None or interface_mode == "vasp":
        import phonopy.interface.vasp as vasp

        vasp.write_vasp(filename, cell)
    elif interface_mode == "abinit":
        import phonopy.interface.abinit as abinit

        abinit.write_abinit(filename, cell)
    elif interface_mode == "qe":
        import phonopy.interface.qe as qe

        pp_filenames = optional_structure_info[1]
        qe.write_pwscf(filename, cell, pp_filenames)

    elif interface_mode == "wien2k":
        import phonopy.interface.wien2k as wien2k

        _, npts, r0s, rmts = optional_structure_info
        wien2k.write_wein2k(filename, cell, npts, r0s, rmts)
    elif interface_mode == "elk":
        import phonopy.interface.elk as elk

        sp_filenames = optional_structure_info[1]
        elk.write_elk(filename, cell, sp_filenames)
    elif interface_mode == "siesta":
        import phonopy.interface.siesta as siesta

        atypes = optional_structure_info[1]
        siesta.write_siesta(filename, cell, atypes)
    elif interface_mode == "cp2k":
        import phonopy.interface.cp2k as cp2k

        _, tree = optional_structure_info
        cp2k.write_cp2k_by_filename(filename, cell, tree)
    elif interface_mode == "crystal":
        import phonopy.interface.crystal as crystal

        conv_numbers = optional_structure_info[1]
        crystal.write_crystal(filename, cell, conv_numbers)
    elif interface_mode == "dftbp":
        import phonopy.interface.dftbp as dftbp

        dftbp.write_dftbp(filename, cell)
    elif interface_mode == "turbomole":
        import phonopy.interface.turbomole as turbomole

        turbomole.write_turbomole(filename, cell)
    elif interface_mode == "aims":
        import phonopy.interface.aims as aims

        aims.write_aims(filename, cell)
    elif interface_mode == "castep":
        import phonopy.interface.castep as castep

        castep.write_castep(filename, cell)
    elif interface_mode == "fleur":
        import phonopy.interface.fleur as fleur

        speci, restlines = optional_structure_info
        fleur.write_fleur(filename, cell, speci, 1, restlines)
    elif interface_mode == "abacus":
        import phonopy.interface.abacus as abacus

        pps = optional_structure_info[1]
        orbitals = optional_structure_info[2]
        abfs = optional_structure_info[3]
        abacus.write_abacus(filename, cell, pps, orbitals, abfs)
    elif interface_mode == "lammps":
        import phonopy.interface.lammps as lammps

        lammps.write_lammps(filename, cell)
    elif interface_mode == "pwmat":
        import phonopy.interface.pwmat as pwmat

        pwmat.write_atom_config(filename, cell)
    else:
        raise RuntimeError("No calculator interface was found.")


def write_supercells_with_displacements(
    interface_mode: str,
    supercell: PhonopyAtoms,
    cells_with_disps: Sequence[PhonopyAtoms],
    optional_structure_info: Optional[tuple] = None,
    displacement_ids: Optional[Union[Sequence, np.ndarray]] = None,
    zfill_width: int = 3,
    additional_info: Optional[dict] = None,
):
    """Write supercell with displacements to files in each calculator format.

    interface_mode : str
        Calculator interface such as 'vasp', 'qe', ...
    supercell : PhonopyAtoms
        Supercell.
    cells_with_disps : list of PhonopyAtoms
        Supercells with displacements.
    optional_structure_info : tuple
        Information returned by the method ``read_crystal_structure``.
        See the docstring.
    displacements_ids : array_like or None, optional
        Integer 1d array with the length of cells_with_disps, containing
        numbers to be assigned to the supercells with displacements.
        Default is None, which gives [1, 2, 3, ...].
    zfill_width : int, optional
        Supercell numbers are filled by zeros from the left with the digits
        as given, which results in 001, 002, ..., when zfill_width=3.
    additional_info : dict or None, optional
        Any information expected to be given to writers of calculators.
        Default is None.

    """
    if displacement_ids is None:
        ids = np.arange(len(cells_with_disps), dtype=int) + 1
    else:
        ids = displacement_ids

    args = (supercell, cells_with_disps, ids)
    kwargs = {"width": zfill_width}
    if additional_info is not None and "pre_filename" in additional_info:
        kwargs["pre_filename"] = additional_info["pre_filename"]

    if interface_mode is None or interface_mode == "vasp":
        import phonopy.interface.vasp as vasp

        vasp.write_supercells_with_displacements(*args, **kwargs)
        write_magnetic_moments(supercell, sort_by_elements=True)
    elif interface_mode == "abinit":
        import phonopy.interface.abinit as abinit

        abinit.write_supercells_with_displacements(*args, **kwargs)
    elif interface_mode == "qe":
        import phonopy.interface.qe as qe

        pp_filenames = optional_structure_info[1]
        qe_args = args + (pp_filenames,)
        qe.write_supercells_with_displacements(*qe_args, **kwargs)
        write_magnetic_moments(supercell, sort_by_elements=False)
    elif interface_mode == "pwmat":
        import phonopy.interface.pwmat as pwmat

        pwmat.write_supercells_with_displacements(*args, **kwargs)

    elif interface_mode == "wien2k":
        import phonopy.interface.wien2k as wien2k

        unitcell_filename, npts, r0s, rmts = optional_structure_info
        N = abs(determinant(additional_info["supercell_matrix"]))
        w2k_args = args + (npts, r0s, rmts, N)
        if "pre_filename" not in kwargs:
            kwargs["pre_filename"] = unitcell_filename
        wien2k.write_supercells_with_displacements(*w2k_args, **kwargs)
    elif interface_mode == "elk":
        import phonopy.interface.elk as elk

        sp_filenames = optional_structure_info[1]
        elk_args = args + (sp_filenames,)
        elk.write_supercells_with_displacements(*elk_args, **kwargs)
    elif interface_mode == "siesta":
        import phonopy.interface.siesta as siesta

        atypes = optional_structure_info[1]
        sst_args = args + (atypes,)
        siesta.write_supercells_with_displacements(*sst_args, **kwargs)
    elif interface_mode == "cp2k":
        import phonopy.interface.cp2k as cp2k

        cp2k_args = args + (optional_structure_info,)
        cp2k.write_supercells_with_displacements(*cp2k_args, **kwargs)
    elif interface_mode == "crystal":
        import phonopy.interface.crystal as crystal

        if additional_info is None:
            kwargs["template_file"] = "TEMPLATE"
        else:
            kwargs["template_file"] = additional_info.get("template_file", "TEMPLATE")
        conv_numbers = optional_structure_info[1]
        N = abs(determinant(additional_info["supercell_matrix"]))
        cst_args = args + (conv_numbers, N)
        crystal.write_supercells_with_displacements(*cst_args, **kwargs)
    elif interface_mode == "dftbp":
        import phonopy.interface.dftbp as dftbp

        dftbp.write_supercells_with_displacements(*args, **kwargs)
    elif interface_mode == "turbomole":
        import phonopy.interface.turbomole as turbomole

        turbomole.write_supercells_with_displacements(*args, **kwargs)
    elif interface_mode == "aims":
        import phonopy.interface.aims as aims

        aims.write_supercells_with_displacements(*args, **kwargs)
    elif interface_mode == "castep":
        import phonopy.interface.castep as castep

        castep.write_supercells_with_displacements(*args, **kwargs)
    elif interface_mode == "fleur":
        import phonopy.interface.fleur as fleur

        speci = optional_structure_info[1]
        restlines = optional_structure_info[2]
        N = abs(determinant(additional_info["supercell_matrix"]))
        fleur_args = args + (speci, N, restlines)
        fleur.write_supercells_with_displacements(*fleur_args, **kwargs)
    elif interface_mode == "abacus":
        import phonopy.interface.abacus as abacus

        pps = optional_structure_info[1]
        orbitals = optional_structure_info[2]
        abfs = optional_structure_info[3]
        abacus_args = args + (pps, orbitals, abfs)
        abacus.write_supercells_with_displacements(*abacus_args, **kwargs)
    elif interface_mode == "lammps":
        import phonopy.interface.lammps as lammps

        lammps.write_supercells_with_displacements(*args, **kwargs)
    else:
        raise RuntimeError("No calculator interface was found.")


def write_magnetic_moments(cell: PhonopyAtoms, sort_by_elements=False):
    """Write MAGMOM."""
    magmoms = cell.magnetic_moments
    if magmoms is not None:
        if sort_by_elements:
            (_, _, _, sort_list) = sort_positions_by_symbols(
                cell.symbols, cell.scaled_positions
            )
        else:
            sort_list = np.arange(len(cell))

        text = " MAGMOM = "
        text += " ".join([f"{v}" for v in magmoms[sort_list].ravel()])
        with open("MAGMOM", "w") as w:
            w.write(text)
            w.write("\n")


def read_crystal_structure(
    filename=None,
    interface_mode=None,
    chemical_symbols=None,
    phonopy_yaml_cls: Optional[type[PhonopyYaml]] = None,
):
    """Return crystal structure from file in each calculator format.

    Parameters
    ----------
    filename : str, optional
        Filename that contains cell structure information. Default is None. The
        predetermined filename for each interface_mode is used.
    interface_mode : str, optional
        This is used to recognize the file format. Default is None, which is
        equivalent to 'vasp' mode.
    chemical_symbols : list of str, optional
        This is only used for 'vasp' mode. VASP POSCAR file format can be
        written without chemical symbol information. With this option, chemical
        symbols can be given.
    phonopy_yaml_cls : PhonopyYaml, optional
        This brings PhonopyYaml-like class dependent parameters. Here, currently
        only the default filenames are provided by this. Default is None.

    Returns
    -------
    tuple
        (Unit cell in PhonopyAtoms, optional_structure_info in tuple)

        The optional_structure_info is given by a tuple. The first element of it
        is the unit cell file name for which the unit cell data are read, and
        the rest is dependent on calculator interface.

    """
    if interface_mode == "phonopy_yaml":
        return _read_phonopy_yaml(filename, phonopy_yaml_cls)

    if filename is None:
        cell_filename = get_default_cell_filename(interface_mode)
        if not pathlib.Path(cell_filename).is_file():
            return None, (cell_filename, "(default file name)")
    else:
        cell_filename = filename
        if not pathlib.Path(cell_filename).is_file():
            return None, (cell_filename,)

    if interface_mode is None or interface_mode == "vasp":
        from phonopy.interface.vasp import read_vasp

        if chemical_symbols is None:
            unitcell = read_vasp(cell_filename)
        else:
            unitcell = read_vasp(cell_filename, symbols=chemical_symbols)
        return unitcell, (cell_filename,)
    elif interface_mode == "abinit":
        from phonopy.interface.abinit import read_abinit

        unitcell = read_abinit(cell_filename)
        return unitcell, (cell_filename,)
    elif interface_mode == "qe":
        from phonopy.interface.qe import read_pwscf

        unitcell, pp_filenames = read_pwscf(cell_filename)
        return unitcell, (cell_filename, pp_filenames)
    elif interface_mode == "pwmat":
        from phonopy.interface.pwmat import read_atom_config

        unitcell = read_atom_config(cell_filename)
        return unitcell, (cell_filename,)
    elif interface_mode == "wien2k":
        from phonopy.interface.wien2k import parse_wien2k_struct

        unitcell, npts, r0s, rmts = parse_wien2k_struct(cell_filename)
        return unitcell, (cell_filename, npts, r0s, rmts)
    elif interface_mode == "elk":
        from phonopy.interface.elk import read_elk

        unitcell, sp_filenames = read_elk(cell_filename)
        return unitcell, (cell_filename, sp_filenames)
    elif interface_mode == "siesta":
        from phonopy.interface.siesta import read_siesta

        unitcell, atypes = read_siesta(cell_filename)
        return unitcell, (cell_filename, atypes)
    elif interface_mode == "cp2k":
        from phonopy.interface.cp2k import read_cp2k

        unitcell, config_tree = read_cp2k(cell_filename)
        return unitcell, (cell_filename, config_tree)
    elif interface_mode == "crystal":
        from phonopy.interface.crystal import read_crystal

        unitcell, conv_numbers = read_crystal(cell_filename)
        return unitcell, (cell_filename, conv_numbers)
    elif interface_mode == "dftbp":
        from phonopy.interface.dftbp import read_dftbp

        unitcell = read_dftbp(cell_filename)
        return unitcell, (cell_filename,)
    elif interface_mode == "turbomole":
        from phonopy.interface.turbomole import read_turbomole

        unitcell = read_turbomole(cell_filename)
        return unitcell, (cell_filename,)
    elif interface_mode == "aims":
        from phonopy.interface.aims import read_aims

        unitcell = read_aims(cell_filename)
        return unitcell, (cell_filename,)
    elif interface_mode == "castep":
        from phonopy.interface.castep import read_castep

        unitcell = read_castep(cell_filename)
        return unitcell, (cell_filename,)
    elif interface_mode == "fleur":
        from phonopy.interface.fleur import read_fleur

        unitcell, speci, restlines = read_fleur(cell_filename)
        return unitcell, (cell_filename, speci, restlines)
    elif interface_mode == "abacus":
        from phonopy.interface.abacus import read_abacus

        unitcell, pps, orbitals, abfs = read_abacus(cell_filename)
        return unitcell, (cell_filename, pps, orbitals, abfs)
    elif interface_mode == "lammps":
        from phonopy.interface.lammps import read_lammps

        unitcell = read_lammps(cell_filename)
        return unitcell, (cell_filename,)
    else:
        raise RuntimeError("No calculator interface was found.")


def get_default_cell_filename(interface_mode):
    """Return default filename of unit cell structure of each calculator."""
    if interface_mode is None or interface_mode == "vasp":
        return "POSCAR"
    elif interface_mode in ("abinit", "qe"):
        return "unitcell.in"
    elif interface_mode == "wien2k":
        return "case.struct"
    elif interface_mode == "elk":
        return "elk.in"
    elif interface_mode == "siesta":
        return "input.fdf"
    elif interface_mode == "cp2k":
        return "unitcell.inp"
    elif interface_mode == "crystal":
        return "crystal.o"
    elif interface_mode == "dftbp":
        return "geo.gen"
    elif interface_mode == "turbomole":
        return "control"
    elif interface_mode == "aims":
        return "geometry.in"
    elif interface_mode == "castep":
        return "unitcell.cell"
    elif interface_mode == "fleur":
        return "fleur.in"
    elif interface_mode == "abacus":
        return "STRU"
    elif interface_mode == "lammps":
        return "unitcell"
    elif interface_mode == "pwmat":
        return "atom.config"
    else:
        return None


def get_default_supercell_filename(interface_mode):
    """Return default filename of supercell structure of each calculator."""
    if interface_mode == "phonopy_yaml":
        return "phonopy_disp.yaml"
    elif interface_mode is None or interface_mode == "vasp":
        return "SPOSCAR"
    elif interface_mode in ("abinit", "elk", "qe", "fleur"):
        return "supercell.in"
    elif interface_mode == "wien2k":
        return "case.structS"
    elif interface_mode == "siesta":
        return "supercell.fdf"
    elif interface_mode == "cp2k":
        # CP2K interface generates filenames based on original project name
        return None
    elif interface_mode == "crystal":
        return None  # supercell.ext can not be parsed by crystal interface.
    elif interface_mode == "dftbp":
        return "geo.genS"
    elif interface_mode == "turbomole":
        return None  # TURBOMOLE interface generates directories with inputs
    elif interface_mode == "aims":
        return "geometry.in.supercell"
    elif interface_mode == "castep":
        return "supercell.cell"
    elif interface_mode == "abacus":
        return "sSTRU"
    elif interface_mode == "lammps":
        return "supercell"
    elif interface_mode == "pwmat":
        return "supercell.config"
    else:
        return None


def get_default_displacement_distance(interface_mode):
    """Return default displacement distance of each calculator."""
    if interface_mode in (
        "wien2k",
        "abinit",
        "elk",
        "qe",
        "siesta",
        "turbomole",
        "fleur",
        "abacus",
    ):
        displacement_distance = 0.02
    else:  # default or vasp, crystal, cp2k, pwmat
        displacement_distance = 0.01
    return displacement_distance


def get_default_physical_units(interface_mode=None) -> dict:
    """Replace get_calculator_get_physical_units()."""
    warnings.warn(
        (
            "get_default_physical_units is deprecated. "
            "Use get_calculator_physical_units instead."
        ),
        DeprecationWarning,
        stacklevel=2,
    )
    return get_calculator_physical_units(interface_mode=interface_mode)


def get_calculator_physical_units(interface_mode=None) -> dict:
    """Return physical units of eachi calculator.

    Physical units: energy,  distance,  atomic mass, force,        force constants
    vasp          : eV,      angstrom,  AMU,         eV/angstrom,  eV/angstrom^2
    wien2k        : Ry,      au(=borh), AMU,         mRy/au,       mRy/au^2
    abinit        : hartree, au,        AMU,         eV/angstrom,  eV/angstrom.au
    elk           : hartree, au,        AMU,         hartree/au,   hartree/au^2
    qe            : Ry,      au,        AMU,         Ry/au,        Ry/au^2
    siesta        : eV,      au,        AMU,         eV/angstrom,  eV/angstrom.au
    CRYSTAL       : eV,      angstrom,  AMU,         eV/angstrom,  eV/angstrom^2
    DFTB+         : hartree, au,        AMU          hartree/au,   hartree/au^2
    TURBOMOLE     : hartree, au,        AMU,         hartree/au,   hartree/au^2
    CP2K          : hartree, angstrom,  AMU,         hartree/au,   hartree/angstrom.au
    FHI-aims      : eV,      angstrom,  AMU,         eV/angstrom,  eV/angstrom^2
    castep        : eV,      angstrom,  AMU,         eV/angstrom,  eV/angstrom^2
    fleur         : hartree, au,        AMU,         hartree/au,   hartree/au^2
    abacus        : eV,      au,        AMU,         eV/angstrom,  eV/angstrom.au
    lammps        : eV,      angstrom,  AMU,         eV/angstrom,  eV/angstrom^2
    pwmat         : eV,      angstrom,  AMU,         eV/angstrom,  eV/angstrom^2

    units['force_constants_unit'] is used in
    the 'get_force_constant_conversion_factor' method.

    """
    units = {
        "factor": None,
        "nac_factor": None,
        "distance_to_A": None,
        "force_to_eVperA": None,
        "force_constants_unit": None,
        "length_unit": None,
        "force_unit": None,
    }
    physical_units = get_physical_units()

    if interface_mode is None or interface_mode in ("vasp", "aims", "lammps", "pwmat"):
        VaspToTHz = physical_units.DefaultToTHz  # [THz] 15.633302
        units["factor"] = VaspToTHz
        units["nac_factor"] = physical_units.Hartree * physical_units.Bohr
        units["distance_to_A"] = 1.0
        units["force_constants_unit"] = "eV/angstrom^2"
        units["length_unit"] = "angstrom"
        units["force_unit"] = "eV/angstrom"
    elif interface_mode == "abinit":
        AbinitToTHz = (
            sqrt(physical_units.EV / (physical_units.AMU * physical_units.Bohr))
            / physical_units.Angstrom
            / (2 * pi)
            / 1e12
        )  # [THz] 21.49068
        units["factor"] = AbinitToTHz
        units["nac_factor"] = physical_units.Hartree / physical_units.Bohr
        units["distance_to_A"] = physical_units.Bohr
        units["force_constants_unit"] = "eV/angstrom.au"
        units["length_unit"] = "au"
        units["force_unit"] = "eV/angstrom"
    elif interface_mode == "qe":
        PwscfToTHz = (
            sqrt(physical_units.Rydberg * physical_units.EV / physical_units.AMU)
            / (physical_units.Bohr * 1e-10)
            / (2 * pi)
            / 1e12
        )  # [THz] 108.97077
        units["factor"] = PwscfToTHz
        units["nac_factor"] = 2.0
        units["distance_to_A"] = physical_units.Bohr
        units["force_to_eVperA"] = physical_units.Rydberg / physical_units.Bohr
        units["force_constants_unit"] = "Ry/au^2"
        units["length_unit"] = "au"
        units["force_unit"] = "Ry/au"
    elif interface_mode == "wien2k":
        Wien2kToTHz = (
            sqrt(physical_units.Rydberg / 1000 * physical_units.EV / physical_units.AMU)
            / (physical_units.Bohr * 1e-10)
            / (2 * pi)
            / 1e12
        )  # [THz] 3.44595837
        units["factor"] = Wien2kToTHz
        units["nac_factor"] = 2000.0
        units["distance_to_A"] = physical_units.Bohr
        units["force_to_eVperA"] = 0.001 * physical_units.Rydberg / physical_units.Bohr
        units["force_constants_unit"] = "mRy/au^2"
        units["length_unit"] = "au"
        units["force_unit"] = "mRy/au"
    elif interface_mode == "elk":
        ElkToTHz = (
            sqrt(physical_units.Hartree * physical_units.EV / physical_units.AMU)
            / (physical_units.Bohr * 1e-10)
            / (2 * pi)
            / 1e12
        )  # [THz] 154.10794
        units["factor"] = ElkToTHz
        units["nac_factor"] = 1.0
        units["distance_to_A"] = physical_units.Bohr
        units["force_constants_unit"] = "hartree/au^2"
        units["length_unit"] = "au"
        units["force_unit"] = "hartree/au"
    elif interface_mode in ["siesta", "abacus"]:
        SiestaToTHz = (
            sqrt(physical_units.EV / (physical_units.AMU * physical_units.Bohr))
            / physical_units.Angstrom
            / (2 * pi)
            / 1e12
        )  # [THz] 21.49068
        units["factor"] = SiestaToTHz
        units["nac_factor"] = physical_units.Hartree / physical_units.Bohr
        units["distance_to_A"] = physical_units.Bohr
        units["force_constants_unit"] = "eV/angstrom.au"
        units["length_unit"] = "au"
        units["force_unit"] = "eV/angstrom"
    elif interface_mode == "cp2k":
        CP2KToTHz = (  # CP2K uses a.u. for forces but Angstrom for distances
            sqrt(
                physical_units.Hartree
                * physical_units.EV
                / (physical_units.AMU * physical_units.Bohr)
            )
            / physical_units.Angstrom
            / (2 * pi)
            / 1e12
        )
        units["factor"] = CP2KToTHz
        units["nac_factor"] = physical_units.Bohr**2
        units["distance_to_A"] = 1.0
        units["force_to_eVperA"] = physical_units.Hartree / physical_units.Bohr
        units["force_constants_unit"] = "hartree/angstrom.au"
        units["length_unit"] = "angstrom"
        units["force_unit"] = "hartree/au"
    elif interface_mode == "crystal":
        CrystalToTHz = physical_units.DefaultToTHz
        units["factor"] = CrystalToTHz
        units["nac_factor"] = physical_units.Hartree * physical_units.Bohr
        units["distance_to_A"] = 1.0
        units["force_constants_unit"] = "eV/angstrom^2"
        units["length_unit"] = "angstrom"
        units["force_unit"] = "eV/angstrom"
    elif interface_mode == "dftbp":
        DftbpToTHz = (
            sqrt(physical_units.Hartree * physical_units.EV / physical_units.AMU)
            / (physical_units.Bohr * 1e-10)
            / (2 * pi)
            / 1e12
        )  # [THz] 154.10794344
        units["factor"] = DftbpToTHz
        units["nac_factor"] = physical_units.Hartree * physical_units.Bohr
        units["distance_to_A"] = physical_units.Bohr
        units["force_constants_unit"] = "hartree/au^2"
        units["length_unit"] = "au"
        units["force_unit"] = "hartree/au"
    elif interface_mode == "turbomole":
        TurbomoleToTHz = (  # Turbomole uses atomic units (Hartree/Bohr)
            sqrt(physical_units.Hartree * physical_units.EV / physical_units.AMU)
            / (physical_units.Bohr * 1e-10)
            / (2 * pi)
            / 1e12
        )  # [THz] 154.10794
        units["factor"] = TurbomoleToTHz
        units["nac_factor"] = 1.0
        units["distance_to_A"] = physical_units.Bohr
        units["force_to_eVperA"] = physical_units.Hartree / physical_units.Bohr
        units["force_constants_unit"] = "hartree/au^2"
        units["length_unit"] = "au"
        units["force_unit"] = "hartree/au"
    elif interface_mode == "castep":
        CastepToTHz = physical_units.DefaultToTHz
        units["factor"] = CastepToTHz
        units["nac_factor"] = physical_units.Hartree * physical_units.Bohr
        units["distance_to_A"] = 1.0
        units["force_constants_unit"] = "eV/angstrom^2"
        units["length_unit"] = "angstrom"
        units["force_unit"] = "eV/angstrom"
    elif interface_mode == "fleur":
        FleurToTHz = (  # Fleur uses atomic units (Hartree/Bohr)
            sqrt(physical_units.Hartree * physical_units.EV / physical_units.AMU)
            / (physical_units.Bohr * 1e-10)
            / (2 * pi)
            / 1e12
        )  # [THz] 154.10794
        units["factor"] = FleurToTHz
        units["nac_factor"] = 1.0
        units["distance_to_A"] = physical_units.Bohr
        units["force_constants_unit"] = "hartree/au^2"
        units["length_unit"] = "au"
        units["force_unit"] = "hartree/au"

    return units


def get_calc_dataset(
    interface_mode: str,
    num_atoms: int,
    force_filenames: str,
    verbose: bool = True,
) -> dict:
    """Read calculator output files and parse force sets.

    Note
    ----
    Wien2k output is treated by ``get_calc_datasets_wien2k``.

    Returns
    -------
    dict:
        "forces": Set of forces in supercells.
        "supercell_energies": Set of supercell energies.

    """
    if interface_mode is None or interface_mode == "vasp":
        from phonopy.interface.vasp import parse_set_of_forces
    elif interface_mode == "abinit":
        from phonopy.interface.abinit import parse_set_of_forces
    elif interface_mode == "qe":
        from phonopy.interface.qe import parse_set_of_forces
    elif interface_mode == "elk":
        from phonopy.interface.elk import parse_set_of_forces
    elif interface_mode == "siesta":
        from phonopy.interface.siesta import parse_set_of_forces
    elif interface_mode == "cp2k":
        from phonopy.interface.cp2k import parse_set_of_forces
    elif interface_mode == "crystal":
        from phonopy.interface.crystal import parse_set_of_forces
    elif interface_mode == "dftbp":
        from phonopy.interface.dftbp import parse_set_of_forces
    elif interface_mode == "turbomole":
        from phonopy.interface.turbomole import parse_set_of_forces
    elif interface_mode == "aims":
        from phonopy.interface.aims import parse_set_of_forces
    elif interface_mode == "castep":
        from phonopy.interface.castep import parse_set_of_forces
    elif interface_mode == "fleur":
        from phonopy.interface.fleur import parse_set_of_forces
    elif interface_mode == "abacus":
        from phonopy.interface.abacus import parse_set_of_forces
    elif interface_mode == "lammps":
        from phonopy.interface.lammps import parse_set_of_forces
    elif interface_mode == "pwmat":
        from phonopy.interface.pwmat import parse_set_of_forces
    else:
        return []

    data_sets = parse_set_of_forces(num_atoms, force_filenames, verbose=verbose)
    if isinstance(data_sets, dict):
        return data_sets
    else:
        return {"forces": data_sets}


def get_calc_dataset_wien2k(
    force_filenames,
    supercell,
    disp_dataset,
    wien2k_P1_mode=False,
    symmetry_tolerance=None,
    verbose=False,
):
    """Read Wien2k output files and parse force sets."""
    from phonopy.interface.wien2k import parse_set_of_forces

    disps, _ = get_displacements_and_forces(disp_dataset)
    force_sets = parse_set_of_forces(
        disps,
        force_filenames,
        supercell,
        wien2k_P1_mode=wien2k_P1_mode,
        symmetry_tolerance=symmetry_tolerance,
        verbose=verbose,
    )
    return {"forces": force_sets}


def get_force_constant_conversion_factor(unit, interface_mode):
    """Return unit conversion factor of force constants."""
    _unit = unit.replace("Angstrom", "angstrom")  # for backward compatibility
    interface_default_units = get_calculator_physical_units(interface_mode)
    default_unit = interface_default_units["force_constants_unit"]
    physical_units = get_physical_units()
    factor_to_eVperA2 = {
        "eV/angstrom^2": 1,
        "eV/angstrom.au": 1 / physical_units.Bohr,
        "Ry/au^2": physical_units.Rydberg / physical_units.Bohr**2,
        "mRy/au^2": physical_units.Rydberg / physical_units.Bohr**2 / 1000,
        "hartree/au^2": physical_units.Hartree / physical_units.Bohr**2,
        "hartree/angstrom.au": physical_units.Hartree / physical_units.Bohr,
    }
    if default_unit not in factor_to_eVperA2:
        msg = "Force constant conversion for %s unit is not implemented."
        raise NotImplementedError(msg)
    if default_unit != _unit:
        factor = factor_to_eVperA2[_unit] / factor_to_eVperA2[default_unit]
        return factor
    else:
        return 1.0


def _read_phonopy_yaml(
    filename: str, phonopy_yaml_cls: Optional[type[PhonopyYaml]]
) -> tuple[PhonopyAtoms, tuple]:
    cell_filename = _get_cell_filename(filename, phonopy_yaml_cls)
    if cell_filename is None:
        return None, (None, None)

    if phonopy_yaml_cls is None:
        phyml = PhonopyYaml()
    else:
        phyml = phonopy_yaml_cls()
    try:
        phyml.read(cell_filename)
    except TypeError:  # yaml.load returns str: File format seems not YAML.
        return None, (cell_filename, None)
    except yaml.parser.ParserError:
        return None, (cell_filename, None)

    cell = phyml.unitcell
    return cell, (cell_filename, phyml)


def _get_cell_filename(
    filename, phonopy_yaml_cls: Optional[type[PhonopyYaml]]
) -> Optional[str]:
    cell_filename = None

    default_filenames = []
    if phonopy_yaml_cls is not None:
        for fname in phonopy_yaml_cls.default_filenames:
            for ext in get_supported_file_extensions_for_compression():
                default_filenames.append(f"{fname}{ext}")

    for fname in [filename] + default_filenames:
        if fname and pathlib.Path(fname).is_file():
            cell_filename = fname
            break
    return cell_filename
